"""Modal subprocess invocation and streaming wrappers."""

from __future__ import annotations

import json
import os
import re
import shlex
import shutil
import subprocess
import sys
import threading
from pathlib import Path
from typing import Any, Dict, Generator, List, Optional

from ..protocol.enums import BackendType
from ..protocol.events import ErrorEvent, LogEvent, OperationCompleteEvent
from ..protocol.models import DeploymentConfig, EndpointInfo, LaunchpadSettings
from .modal_auth import get_modal_profile
from .naming import default_served_model_name
from .naming import infer_backend_from_app_name, infer_instance_from_app_name, legacy_app_name
from .naming import modal_function_name


class ModalBackend:
    """Wrapper around Modal CLI subprocess calls."""

    # ------------------------------------------------------------------
    # Subprocess lifecycle tracking
    # ------------------------------------------------------------------

    _active_procs: set[subprocess.Popen] = set()
    _active_procs_lock = threading.Lock()
    _shutdown_event = threading.Event()
    _CLI_TIMEOUT_SECONDS = 8.0

    @staticmethod
    def modal_cli_path() -> Optional[str]:
        """Resolve the Modal CLI, preferring the active environment's scripts dir."""
        env_prefix = Path(sys.prefix)
        candidates = [
            env_prefix / "bin" / "modal",
            env_prefix / "Scripts" / "modal.exe",
        ]
        for candidate in candidates:
            if candidate.exists():
                return str(candidate)
        return shutil.which("modal")

    @classmethod
    def _resolve_command(cls, command: List[str]) -> List[str]:
        if command and command[0] == "modal":
            modal_cli = cls.modal_cli_path()
            if modal_cli:
                return [modal_cli, *command[1:]]
        return command

    @classmethod
    def register_proc(cls, proc: subprocess.Popen) -> None:
        """Track a subprocess so it can be terminated on shutdown."""
        with cls._active_procs_lock:
            cls._active_procs.add(proc)

    @classmethod
    def unregister_proc(cls, proc: subprocess.Popen) -> None:
        """Remove a subprocess from tracking."""
        with cls._active_procs_lock:
            cls._active_procs.discard(proc)

    @classmethod
    def terminate_all(cls) -> None:
        """Signal shutdown and terminate all tracked subprocesses.

        Called on app exit to unblock any worker threads that are waiting
        on subprocess I/O so Python can shut down cleanly.
        """
        cls._shutdown_event.set()
        import time
        time.sleep(0.05)
        with cls._active_procs_lock:
            procs = list(cls._active_procs)
            cls._active_procs.clear()
        for proc in procs:
            try:
                proc.terminate()
                if proc.stdout:
                    proc.stdout.close()
                if proc.stderr and proc.stderr != subprocess.PIPE:
                    proc.stderr.close()
            except Exception:
                pass

    @classmethod
    def is_shutting_down(cls) -> bool:
        """Return True if a shutdown has been requested."""
        return cls._shutdown_event.is_set()

    # ------------------------------------------------------------------
    # Pre-flight checks
    # ------------------------------------------------------------------

    @staticmethod
    def is_cli_available() -> bool:
        return ModalBackend.modal_cli_path() is not None

    @staticmethod
    def get_username() -> Optional[str]:
        """Return the current Modal profile/workspace slug, or None."""
        return get_modal_profile()

    # ------------------------------------------------------------------
    # URL helpers
    # ------------------------------------------------------------------

    @staticmethod
    def default_server_url(
        username: str,
        backend: Optional[BackendType] = None,
        app_name: Optional[str] = None,
        function_slug: Optional[str] = None,
    ) -> str:
        resolved = app_name or legacy_app_name(backend or BackendType.LLAMACPP)
        serve_name = modal_function_name("serve", function_slug)
        return f"https://{username}--{resolved}-{serve_name}.modal.run"

    @staticmethod
    def extract_modal_web_url(line: str) -> Optional[str]:
        """Extract the first Modal web URL from an output line."""
        match = re.search(r"https://[A-Za-z0-9-]+\.modal\.run", line or "")
        return match.group(0) if match else None

    @staticmethod
    def test_curl_command(
        backend: BackendType,
        server_url: str,
        served_model_name: Optional[str] = None,
    ) -> str:
        base = server_url.rstrip("/")
        content = "Say hello in one short sentence."

        def _curl_json(endpoint: str, payload: Dict[str, Any]) -> str:
            body = shlex.quote(json.dumps(payload, separators=(",", ":")))
            return (
                f"curl -s -X POST {base}{endpoint} "
                "-H 'Content-Type: application/json' "
                f"-d {body}"
            )

        if backend == BackendType.VLLM:
            model = (served_model_name or "").strip() or os.environ.get(
                "SERVED_MODEL_NAME",
                default_served_model_name(os.environ.get("MODEL_NAME")),
            )
            return _curl_json(
                "/v1/chat/completions",
                {
                    "model": model,
                    "messages": [{"role": "user", "content": content}],
                },
            )
        model = (served_model_name or "").strip() or "default"
        return _curl_json(
            "/v1/completions",
            {
                "model": model,
                "prompt": content,
                "max_tokens": 32,
            },
        )

    # ------------------------------------------------------------------
    # Environment builders
    # ------------------------------------------------------------------

    @staticmethod
    def env_for_backend(config: DeploymentConfig) -> Dict[str, str]:
        """Derive backend-specific env vars from a deployment config."""
        env: Dict[str, str] = {}
        if config.app_name:
            env["MODAL_APP_NAME"] = config.app_name
        if config.function_slug:
            env["MODAL_FUNCTION_SLUG"] = config.function_slug
        if config.backend == BackendType.LLAMACPP:
            if config.llamacpp_image_no_cache is not None:
                env["LLAMA_CPP_IMAGE_NO_CACHE"] = "true" if config.llamacpp_image_no_cache else "false"
            return env
        if config.backend != BackendType.VLLM:
            return env
        if config.model_name:
            env["MODEL_NAME"] = config.model_name
        if config.model_revision:
            env["MODEL_REVISION"] = config.model_revision
        if config.served_model_name:
            env["SERVED_MODEL_NAME"] = config.served_model_name
        if config.fast_boot is not None:
            env["FAST_BOOT"] = "true" if config.fast_boot else "false"
        if config.n_gpu is not None and config.n_gpu > 0:
            env["N_GPU"] = str(config.n_gpu)
        if config.trust_remote_code is not None:
            env["TRUST_REMOTE_CODE"] = "true" if config.trust_remote_code else "false"
        if config.reasoning_parser:
            env["REASONING_PARSER"] = config.reasoning_parser
        if config.tool_call_parser:
            env["TOOL_CALL_PARSER"] = config.tool_call_parser
        if config.default_chat_template_kwargs:
            env["DEFAULT_CHAT_TEMPLATE_KWARGS"] = config.default_chat_template_kwargs
        return env

    @staticmethod
    def build_full_env(
        settings: LaunchpadSettings,
        config: DeploymentConfig,
    ) -> Dict[str, str]:
        env = settings.to_env()
        gpu_type = (config.gpu_type or "").strip()
        if gpu_type and config.gpu_count is not None and config.gpu_count > 0:
            env["GPU_CONFIG"] = f"{gpu_type.upper()}:{config.gpu_count}"
        env.update(ModalBackend.env_for_backend(config))
        return env

    # ------------------------------------------------------------------
    # Command builders
    # ------------------------------------------------------------------

    @staticmethod
    def build_deploy_command(backend: BackendType, app_name: Optional[str] = None) -> List[str]:
        cmd = ["modal", "deploy", backend.script]
        if app_name:
            cmd += ["--name", app_name]
        return cmd

    @staticmethod
    def build_run_command(config: DeploymentConfig) -> List[str]:
        if config.backend == BackendType.VLLM:
            return ["modal", "run", BackendType.VLLM.script]

        args: List[str] = ["modal", "run", f"{BackendType.LLAMACPP.script}::main"]
        if config.preset:
            args += ["--preset", config.preset]
        if config.repo_id:
            args += ["--repo-id", config.repo_id]
        if config.quant:
            args += ["--quant", config.quant]
        if config.revision:
            args += ["--revision", config.revision]
        if config.preload:
            args += ["--preload"]
        else:
            args += ["--no-preload"]
        if config.do_deploy:
            args += ["--deploy"]
        if config.server_args:
            args.append(f"--server-args={config.server_args}")
        if config.host:
            args += ["--host", config.host]
        if config.port is not None:
            args += ["--port", str(config.port)]
        if config.n_gpu_layers is not None:
            args += ["--n-gpu-layers", str(config.n_gpu_layers)]
        return args

    @staticmethod
    def build_modal_entrypoint_command(
        script: str,
        entrypoint: str,
        args: Optional[List[str]] = None,
    ) -> List[str]:
        cmd = ["modal", "run", f"{script}::{entrypoint}"]
        if args:
            cmd.extend(args)
        return cmd

    # ------------------------------------------------------------------
    # Log follow helper
    # ------------------------------------------------------------------

    @staticmethod
    def logs_follow_args() -> List[str]:
        """Best available follow flag for ``modal app logs``."""
        try:
            res = subprocess.run(
                ModalBackend._resolve_command(["modal", "app", "logs", "--help"]),
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                text=True,
                timeout=ModalBackend._CLI_TIMEOUT_SECONDS,
            )
            text = (res.stdout or "") + (res.stderr or "")
            if "--follow" in text:
                return ["--follow"]
            if "-f" in text:
                return ["-f"]
        except subprocess.TimeoutExpired:
            return []
        except Exception:
            pass
        return []

    # ------------------------------------------------------------------
    # App listing
    # ------------------------------------------------------------------

    @staticmethod
    def list_apps() -> Optional[List[EndpointInfo]]:
        """Return parsed app list (possibly empty), or None on failure."""
        try:
            result = subprocess.run(
                ModalBackend._resolve_command(["modal", "app", "list", "--json"]),
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                text=True,
                timeout=ModalBackend._CLI_TIMEOUT_SECONDS,
            )
        except subprocess.TimeoutExpired:
            return None
        except Exception:
            return None
        if result.returncode != 0:
            return None
        try:
            payload: Any = json.loads(result.stdout or "[]")
        except Exception:
            return None

        return _extract_modal_app_rows(payload)

    @staticmethod
    def list_apps_raw() -> Optional[str]:
        """Fallback: return raw text from ``modal app list``."""
        try:
            result = subprocess.run(
                ModalBackend._resolve_command(["modal", "app", "list"]),
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                text=True,
                timeout=ModalBackend._CLI_TIMEOUT_SECONDS,
            )
            if result.returncode == 0:
                return result.stdout or ""
        except subprocess.TimeoutExpired:
            return None
        except Exception:
            pass
        return None

    @staticmethod
    def billing_report_json() -> tuple[Optional[Any], Optional[str]]:
        """Return billing report JSON and an optional error message."""
        commands = [
            ["modal", "billing", "report", "--for", "this month", "--json"],
            ["modal", "workspace", "billing", "report", "--for", "this month", "--json"],
        ]

        def _looks_like_unsupported_command(message: str) -> bool:
            normalized = (message or "").lower()
            return (
                "no such command" in normalized
                or "invalid choice" in normalized
                or "unrecognized arguments" in normalized
                or "usage: modal [options] command" in normalized
            )

        last_error: Optional[str] = None
        unsupported_command_seen = False
        for command in commands:
            try:
                result = subprocess.run(
                    ModalBackend._resolve_command(command),
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                    text=True,
                    timeout=ModalBackend._CLI_TIMEOUT_SECONDS,
                )
            except subprocess.TimeoutExpired:
                return None, "Timed out while requesting billing report."
            except FileNotFoundError:
                return None, "Modal CLI not found in PATH."
            except Exception as exc:
                return None, str(exc)

            if result.returncode == 0:
                try:
                    return json.loads(result.stdout or "{}"), None
                except Exception:
                    return None, "Billing report returned invalid JSON."

            stderr = (result.stderr or "").strip()
            stdout = (result.stdout or "").strip()
            details = stderr or stdout or f"Exit code {result.returncode}"
            last_error = details
            if _looks_like_unsupported_command(details):
                unsupported_command_seen = True
                continue
            return None, details

        if unsupported_command_seen:
            return (
                None,
                "Billing report command is unavailable in this Modal CLI version. "
                "Upgrade Modal CLI and retry.",
            )

        return None, last_error or "Could not read billing report."

    @staticmethod
    def list_volume(volume_name: str, path: str = "/") -> Optional[List[Dict[str, Any]]]:
        """List files/directories in a Modal Volume path."""
        if ModalBackend.is_shutting_down():
            return []
        try:
            result = subprocess.run(
                ModalBackend._resolve_command(["modal", "volume", "ls", volume_name, path, "--json"]),
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                text=True,
                timeout=ModalBackend._CLI_TIMEOUT_SECONDS,
            )
        except subprocess.TimeoutExpired:
            return None
        except Exception:
            return None
        if result.returncode != 0:
            return None
        try:
            payload = json.loads(result.stdout or "[]")
        except Exception:
            return None
        if isinstance(payload, list):
            return [entry for entry in payload if isinstance(entry, dict)]
        if isinstance(payload, dict):
            nested = payload.get("entries") or payload.get("items") or []
            if isinstance(nested, list):
                return [entry for entry in nested if isinstance(entry, dict)]
        return None

    # ------------------------------------------------------------------
    # Streaming subprocess execution
    # ------------------------------------------------------------------

    @staticmethod
    def run_streaming(
        command: List[str],
        env: Optional[Dict[str, str]] = None,
    ) -> Generator[LogEvent | OperationCompleteEvent | ErrorEvent, None, None]:
        """Run *command* and yield protocol events for each output line."""
        merged = {**os.environ, **(env or {})}
        try:
            proc = subprocess.Popen(
                ModalBackend._resolve_command(command),
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                text=True,
                env=merged,
                bufsize=1,
            )
        except Exception as exc:
            yield ErrorEvent(message=str(exc), exit_code=1, recoverable=False)
            return

        ModalBackend.register_proc(proc)
        try:
            assert proc.stdout is not None
            for raw_line in proc.stdout:
                yield LogEvent(line=raw_line.rstrip("\n"))

            proc.wait()
            if proc.returncode == 0:
                yield OperationCompleteEvent(success=True, exit_code=0)
            else:
                yield OperationCompleteEvent(
                    success=False,
                    exit_code=proc.returncode,
                    detail=f"Process exited with code {proc.returncode}",
                )
        finally:
            ModalBackend.unregister_proc(proc)

    @staticmethod
    def run_modal_script_entrypoint(
        script: str,
        entrypoint: str,
        args: Optional[List[str]] = None,
        env: Optional[Dict[str, str]] = None,
    ) -> Generator[LogEvent | OperationCompleteEvent | ErrorEvent, None, None]:
        """Run `modal run <script>::<entrypoint>` and stream output."""
        cmd = ModalBackend.build_modal_entrypoint_command(script, entrypoint, args=args)
        yield from ModalBackend.run_streaming(cmd, env=env)

    @staticmethod
    def run_modal_script_entrypoint_capture(
        script: str,
        entrypoint: str,
        args: Optional[List[str]] = None,
        env: Optional[Dict[str, str]] = None,
    ) -> Optional[tuple[int, str, str]]:
        """Run `modal run <script>::<entrypoint>` and capture stdout/stderr."""
        cmd = ModalBackend.build_modal_entrypoint_command(script, entrypoint, args=args)
        merged = {**os.environ, **(env or {})}
        try:
            result = subprocess.run(
                ModalBackend._resolve_command(cmd),
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                text=True,
                env=merged,
                timeout=ModalBackend._CLI_TIMEOUT_SECONDS,
            )
        except subprocess.TimeoutExpired:
            return None
        except Exception:
            return None
        return result.returncode, result.stdout or "", result.stderr or ""

    @staticmethod
    def run_volume_remove(
        volume_name: str,
        remote_path: str,
        recursive: bool = True,
    ) -> Generator[LogEvent | OperationCompleteEvent | ErrorEvent, None, None]:
        """Delete a file/directory from a Modal Volume."""
        cmd = ["modal", "volume", "rm", volume_name, remote_path]
        if recursive:
            cmd.append("--recursive")
        yield from ModalBackend.run_streaming(cmd)

    @staticmethod
    def run_blocking(
        command: List[str],
        env: Optional[Dict[str, str]] = None,
    ) -> int:
        """Run *command* blocking and return exit code."""
        merged = None
        if env:
            merged = {**os.environ, **env}
        result = subprocess.run(ModalBackend._resolve_command(command), text=True, env=merged)
        return result.returncode


# ------------------------------------------------------------------
# Internal helpers
# ------------------------------------------------------------------

def _non_empty_text(value: Any) -> Optional[str]:
    if value is None:
        return None
    text = str(value).strip()
    return text or None


def _nested_string_for_keys(
    value: Any,
    keys: set[str],
    *,
    max_depth: int = 6,
) -> Optional[str]:
    """Depth-limited search for a string value by key name (case-insensitive)."""
    if max_depth < 0:
        return None

    if isinstance(value, dict):
        for key, child in value.items():
            if str(key).strip().lower() in keys:
                text = _non_empty_text(child)
                if text:
                    return text
        for child in value.values():
            found = _nested_string_for_keys(child, keys, max_depth=max_depth - 1)
            if found:
                return found
        return None

    if isinstance(value, list):
        for child in value:
            found = _nested_string_for_keys(child, keys, max_depth=max_depth - 1)
            if found:
                return found
        return None

    return None


def _collect_modal_web_urls(value: Any, *, max_depth: int = 6) -> List[str]:
    """Collect Modal web URLs found anywhere in a nested JSON-like payload."""
    if max_depth < 0:
        return []

    urls: List[str] = []
    if isinstance(value, dict):
        for child in value.values():
            urls.extend(_collect_modal_web_urls(child, max_depth=max_depth - 1))
        return urls

    if isinstance(value, list):
        for child in value:
            urls.extend(_collect_modal_web_urls(child, max_depth=max_depth - 1))
        return urls

    if isinstance(value, str):
        maybe_url = ModalBackend.extract_modal_web_url(value)
        if maybe_url:
            urls.append(maybe_url)
    return urls


def _extract_item_web_url(item: dict[str, Any]) -> Optional[str]:
    """Best-effort extraction of a Modal web URL from one app-list row."""
    for key in ("web_url", "webUrl", "url", "URL"):
        candidate = item.get(key)
        if isinstance(candidate, str):
            maybe_url = ModalBackend.extract_modal_web_url(candidate)
            if maybe_url:
                return maybe_url

    urls = _collect_modal_web_urls(item)
    if not urls:
        return None

    unique_urls = list(dict.fromkeys(urls))
    unique_urls.sort(key=lambda url: (url.endswith("-dev.modal.run"), len(url)))
    return unique_urls[0]


def _extract_modal_app_rows(payload: Any) -> List[EndpointInfo]:
    rows: List[EndpointInfo] = []
    if isinstance(payload, dict):
        if isinstance(payload.get("apps"), list):
            payload = payload["apps"]
        elif isinstance(payload.get("data"), list):
            payload = payload["data"]
    if not isinstance(payload, list):
        return rows

    for item in payload:
        if not isinstance(item, dict):
            continue
        name = str(
            item.get("name")
            or item.get("app_name")
            or item.get("app")
            or item.get("label")
            or item.get("description")
            or item.get("Description")
            or ""
        ).strip()
        app_id = str(
            item.get("app_id")
            or item.get("id")
            or item.get("appId")
            or item.get("App ID")
            or ""
        ).strip()
        state = str(
            item.get("state")
            or item.get("status")
            or item.get("phase")
            or item.get("State")
            or "unknown"
        ).strip()
        served_model_name = _nested_string_for_keys(item, {"served_model_name"})
        model_name = _nested_string_for_keys(item, {"model_name"})
        repo_id = _nested_string_for_keys(item, {"repo_id", "model_repo_id"})
        quant = _nested_string_for_keys(item, {"quant"})
        backend = infer_backend_from_app_name(name)
        if backend is None:
            if model_name:
                backend = BackendType.VLLM
            elif repo_id or quant:
                backend = BackendType.LLAMACPP
        instance = infer_instance_from_app_name(name, backend)
        rows.append(
            EndpointInfo(
                name=name,
                app_id=app_id,
                state=state,
                backend=backend,
                instance_name=instance,
                web_url=_extract_item_web_url(item),
                served_model_name=served_model_name,
                model_name=model_name,
                repo_id=repo_id,
                quant=quant,
            )
        )
    return rows



"""Deploy log summarization for a concise TUI deploy/warmup view."""

from __future__ import annotations

import re

from ..protocol.enums import BackendType, OperationType

_ERROR_WORD_RE = re.compile(
    r"\b(error|exception|traceback|fatal|failed|runtimeerror|assertionerror)\b",
    flags=re.IGNORECASE,
)
_OOM_RE = re.compile(
    r"(cuda out of memory|outofmemory|segmentation fault|\boom\b)",
    flags=re.IGNORECASE,
)
_HTTP_ERROR_CONTEXT_RE = re.compile(
    r"(\bHTTP\s*5\d\d\b|\b->\s*5\d\d\b|\b5\d\d\s+(Service Unavailable|Bad Gateway|Internal Server Error)\b)",
    flags=re.IGNORECASE,
)
_VLLM_INFO_LINE_RE = re.compile(r"^\((APIServer|EngineCore[^)]*) pid=\d+\)\s+INFO\b")
_LLAMACPP_DOWNLOAD_INFLIGHT_RE = re.compile(r"\binflight=(\d+)\b")
_LLAMACPP_DOWNLOAD_RATE_RE = re.compile(r"\bavg_rate=([0-9]+(?:\.[0-9]+)?)MiB/s\b")
_LLAMACPP_DOWNLOAD_PCT_RE = re.compile(r"\bpct=(\d+)%")


class DeployLogSummarizer:
    """Convert raw backend logs into canonical milestone messages."""

    def __init__(self, backend: BackendType) -> None:
        self.backend = backend
        self.reset()

    def reset(self) -> None:
        self._operation: OperationType | None = None
        self._seen_milestones: set[str] = set()

    def transform(self, line: str, operation: OperationType | None) -> list[str]:
        text = (line or "").rstrip()
        if not text:
            return []

        if operation != self._operation:
            self._operation = operation
            self._seen_milestones.clear()

        preserved = self._preserve_common_line(text)
        if preserved is not None:
            return preserved

        if self.backend == BackendType.LLAMACPP and self._is_llamacpp_readiness_503_noise(text):
            return []

        mapped = self._map_backend_line(text)
        if mapped:
            return self._emit_once(mapped)

        if self._is_error_like(text):
            return [text]

        if self._is_noise(text):
            return []

        return []

    def _emit_once(self, milestone: str) -> list[str]:
        if milestone in self._seen_milestones:
            return []
        self._seen_milestones.add(milestone)
        return [milestone]

    def _preserve_common_line(self, text: str) -> list[str] | None:
        stripped = text.strip()
        if text.startswith("Running: "):
            return [text]
        if stripped.startswith("env: "):
            return [text]
        if text.startswith("Probing readiness at: "):
            out = self._emit_once("Waiting for readiness")
            out.append(text)
            return out
        if text.startswith("Waiting for GPU scheduling"):
            return [text]
        if text == "GPU allocated; continuing readiness probe.":
            return [text]
        if text == "Server is ready!":
            return [text]
        if text.startswith("Test command:"):
            return [text]
        if text.startswith("=== OpenAI-compatible"):
            return [text]
        if text == "=========================":
            return [text]
        if text == "OpenAI-compatible connection summary:":
            return [text]
        if text.startswith("OpenCode custom provider:"):
            return [text]
        if stripped.startswith("Base URL: "):
            return [text]
        if stripped.startswith("Model ID: "):
            return [text]
        if stripped.startswith("Provider ID: "):
            return [text]
        if stripped.startswith("Display name: "):
            return [text]
        if stripped.startswith("API key: "):
            return [text]
        if stripped == "Models:":
            return [text]
        if stripped.startswith("- ID: "):
            return [text]
        if stripped.startswith("Name: "):
            return [text]
        if stripped.startswith("Headers: "):
            return [text]
        if "Created web function serve =>" in text and "http" in text:
            if "-dev.modal.run" in text:
                return []
            return [text]
        if text.startswith("View Deployment: "):
            return [text]
        if "App deployed in " in text:
            return [text]
        if text.startswith("Warning: "):
            return [text]
        return None

    def _map_backend_line(self, text: str) -> str | None:
        if self.backend == BackendType.LLAMACPP:
            return self._map_llamacpp_line(text)
        return self._map_vllm_line(text)

    def _map_llamacpp_line(self, text: str) -> str | None:
        if "cache hit:" in text:
            return "Loading cached model"
        if "cache miss" in text:
            return "Downloading model"
        # The preload helper prints a "downloading ..." banner unconditionally
        # before calling snapshot_download(), even for cache hits. Only show the
        # canonical download milestone once we see progress output.
        if "download in progress..." in text:
            if self._llamacpp_progress_indicates_actual_transfer(text):
                pct = self._llamacpp_progress_percent(text)
                if pct is not None:
                    return f"Downloading model ({pct}%)"
                return "Downloading model"
            return None
        if "found GGUF entries:" in text:
            return "Loading cached model"
        if "using model file:" in text:
            return "Loading cached model"
        if "starting llama-server:" in text:
            return "Starting server"
        if text.startswith("ggml_cuda_init:") or text.startswith("load_backend:"):
            return "Initializing CUDA"
        if (
            text.startswith("main: loading model")
            or text.startswith("srv    load_model:")
            or text.startswith("llama_model_loader: loaded meta data")
        ):
            return "Loading model metadata"
        if text.startswith("load_tensors: loading model tensors"):
            return "Loading model weights"
        if text.startswith("load_tensors: offloading") or text.startswith("load_tensors: offloaded"):
            return "Loading weights on GPU"
        return None

    def _map_vllm_line(self, text: str) -> str | None:
        if text.startswith("Starting vLLM command:"):
            return "Starting server"
        if "vLLM API server version" in text:
            return "Starting server"
        if "Resolved architecture:" in text:
            return "Loading model metadata"
        if "Starting to load model " in text:
            return "Loading model weights"
        if "Time spent downloading weights" in text:
            return "Downloading model"
        if text.startswith("Loading safetensors checkpoint shards:"):
            return "Loading model weights"
        if "Model loading took " in text:
            return "Loading weights on GPU"
        if "Capturing CUDA graphs (" in text:
            return "Capturing CUDA graphs"
        if "Available KV cache memory" in text or "GPU KV cache size:" in text:
            return "Initializing KV cache"
        if self._is_vllm_compile_line(text):
            return "Compiling kernels"
        return None

    @staticmethod
    def _is_vllm_compile_line(text: str) -> bool:
        if "[monitor.py:" in text and "torch.compile" in text:
            return True
        if "[backends.py:" not in text:
            return False
        return (
            "torch.compile" in text
            or "Dynamo bytecode transform time" in text
            or "compiled graph" in text
        )

    def _is_noise(self, text: str) -> bool:
        if self.backend == BackendType.LLAMACPP:
            return self._is_llamacpp_noise(text)
        return self._is_vllm_noise(text)

    @staticmethod
    def _is_llamacpp_noise(text: str) -> bool:
        stripped = text.strip()
        return (
            "download in progress..." in text
            or text.startswith("llama_model_loader: - kv")
            or text.startswith("llama_model_loader: - type")
            or text.startswith("print_info:")
            or text.startswith("load: ")
            or text.startswith("system info:")
            or text.startswith("system_info:")
            or text.startswith("srv  log_server_r:")
            or text.startswith("common_init_result:")
            or text.startswith("llama_params_fit")
            or text.startswith("llama_params_fit_impl:")
            or text.startswith("llama_model_load_from_file_impl:")
            or text.startswith("load_tensors:")
            or text.startswith("main: n_parallel")
            or text.startswith("build:")
            or text.startswith("Running without SSL")
            or text.startswith("init: using ")
            or text.startswith("start: binding port")
            or stripped == "Next steps:"
            or stripped.startswith("Use the exact URL from the `Created web function serve =>")
            or stripped.startswith("(Modal may truncate long labels and append a hash")
            or stripped.startswith("1) Deploy the server:")
            or stripped.startswith("2) Once deployed, curl the server")
            or stripped.startswith("3) Faster iteration")
            or stripped.startswith("4) Image cache behavior")
            or stripped.startswith("Default: reuse cached image layers")
            or stripped.startswith("Force fresh latest image pull/build:")
            or stripped.startswith("Example: LLAMA_CPP_IMAGE_NO_CACHE=true modal deploy")
            or stripped.startswith("modal run ")
            or stripped.startswith("modal deploy ")
            or stripped.startswith("curl -sS -X POST ")
            or stripped.startswith("✅ Weights cached in Modal Volume")
            or stripped.startswith("ℹ️ `--deploy` runs `modal deploy`")
            or stripped.startswith("🚀 Deploying...")
            or stripped.startswith("✅ Deploy triggered.")
        )

    @staticmethod
    def _is_vllm_noise(text: str) -> bool:
        if _VLLM_INFO_LINE_RE.match(text):
            return True
        if text.lstrip().startswith("vllm serve "):
            return True
        if text.startswith("Loading safetensors checkpoint shards:"):
            return True
        if "Capturing CUDA graphs (" in text:
            return True
        return False

    @staticmethod
    def _is_llamacpp_readiness_503_noise(text: str) -> bool:
        if "/v1/completions" not in text or "503" not in text:
            return False
        return text.startswith("srv  log_server_r:") or text.lstrip().startswith("POST /v1/completions")

    @staticmethod
    def _is_error_like(text: str) -> bool:
        if _ERROR_WORD_RE.search(text) or _OOM_RE.search(text):
            return True
        if _HTTP_ERROR_CONTEXT_RE.search(text):
            return True
        return False

    @staticmethod
    def _llamacpp_progress_indicates_actual_transfer(text: str) -> bool:
        inflight_match = _LLAMACPP_DOWNLOAD_INFLIGHT_RE.search(text)
        if inflight_match and int(inflight_match.group(1)) > 0:
            return True
        rate_match = _LLAMACPP_DOWNLOAD_RATE_RE.search(text)
        if rate_match:
            try:
                return float(rate_match.group(1)) > 0.0
            except ValueError:
                return False
        return False

    @staticmethod
    def _llamacpp_progress_percent(text: str) -> int | None:
        pct_match = _LLAMACPP_DOWNLOAD_PCT_RE.search(text)
        if pct_match is None:
            return None
        try:
            pct = int(pct_match.group(1))
        except ValueError:
            return None
        return max(0, min(100, pct))



"""Protocol layer: pure data models, enums, and event contracts.

This package is shared by Core, TUI, and CLI layers. It has zero UI
dependencies and defines the strict communication contract.
"""

from .enums import BackendType, DeploymentState, OperationType
from .events import (
    BaseEvent,
    ErrorEvent,
    LogEvent,
    OperationCompleteEvent,
    StateChangeEvent,
)
from .models import (
    DeploymentConfig,
    EndpointInfo,
    LaunchpadSettings,
    StoredModelInfo,
    StorageSnapshot,
)

__all__ = [
    "BackendType",
    "BaseEvent",
    "DeploymentConfig",
    "DeploymentState",
    "EndpointInfo",
    "ErrorEvent",
    "LaunchpadSettings",
    "LogEvent",
    "OperationCompleteEvent",
    "OperationType",
    "StateChangeEvent",
    "StoredModelInfo",
    "StorageSnapshot",
]



"""Reusable form input helpers for deploy/settings screens."""

from __future__ import annotations

from textual.app import ComposeResult
from textual.containers import Vertical
from textual.widgets import Input, Static, Switch
from textual.validation import Length


class FormField(Vertical):
    """A labelled input field with optional hint text."""

    DEFAULT_CSS = """
    FormField {
        height: auto;
        padding: 0 0 1 0;
    }
    FormField .form-label {
        color: #c8d6c9;
        height: 1;
    }
    FormField .form-hint {
        color: #7f9082;
        height: 1;
    }
    """

    def __init__(
        self,
        label: str,
        field_id: str,
        default: str = "",
        hint: str = "",
        password: bool = False,
        required: bool = False,
        **kwargs: object,
    ) -> None:
        super().__init__(**kwargs)
        self._label = label
        self._field_id = field_id
        self._default = default
        self._hint = hint
        self._password = password
        self._required = required

    def compose(self) -> ComposeResult:
        yield Static(self._label, classes="form-label")
        validators = [Length(minimum=1)] if self._required else []
        placeholder = "" if self._hint else self._label
        yield Input(
            value=self._default,
            placeholder=placeholder,
            password=self._password,
            id=self._field_id,
            validators=validators,
        )
        if self._hint:
            yield Static(self._hint, classes="form-hint")

    @property
    def value(self) -> str:
        return self.query_one(f"#{self._field_id}", Input).value


class ToggleField(Vertical):
    """A labelled boolean toggle (switch)."""

    DEFAULT_CSS = """
    ToggleField {
        height: auto;
        padding: 0 0 1 0;
    }
    ToggleField .form-label {
        color: #c8d6c9;
        height: 1;
    }
    """

    def __init__(
        self,
        label: str,
        field_id: str,
        default: bool = False,
        **kwargs: object,
    ) -> None:
        super().__init__(**kwargs)
        self._label = label
        self._field_id = field_id
        self._default = default

    def compose(self) -> ComposeResult:
        yield Static(self._label, classes="form-label")
        yield Switch(value=self._default, id=self._field_id)

    @property
    def value(self) -> bool:
        return self.query_one(f"#{self._field_id}", Switch).value



"""Worker bridge: Core orchestrator generators -> Textual messages.

Each worker runs a Core generator in a thread and posts protocol events
as Textual messages to the app/screen.
"""

from __future__ import annotations

from textual.message import Message
from ..core.hf_models import ModelCandidate

from ..protocol.enums import DeploymentState, OperationType
from ..protocol.events import (
    BaseEvent,
    ErrorEvent,
    LogEvent,
    OperationCompleteEvent,
    StateChangeEvent,
)
from ..protocol.models import StorageSnapshot


# -----------------------------------------------------------------------
# Textual Messages (thin wrappers around protocol events)
# -----------------------------------------------------------------------


class LogMessage(Message):
    """A log line from an operation."""

    def __init__(self, line: str, stream: str = "stdout") -> None:
        super().__init__()
        self.line = line
        self.stream = stream


class StateChanged(Message):
    """Deployment state transition."""

    def __init__(
        self,
        state: DeploymentState,
        operation: OperationType | None = None,
        detail: str = "",
    ) -> None:
        super().__init__()
        self.state = state
        self.operation = operation
        self.detail = detail


class OperationDone(Message):
    """An operation finished."""

    def __init__(
        self,
        operation: OperationType,
        success: bool,
        exit_code: int = 0,
        detail: str = "",
        data: object = None,
    ) -> None:
        super().__init__()
        self.operation = operation
        self.success = success
        self.exit_code = exit_code
        self.detail = detail
        self.data = data


class OperationError(Message):
    """An error occurred."""

    def __init__(self, message: str, recoverable: bool = True) -> None:
        super().__init__()
        self.message = message
        self.recoverable = recoverable


class VllmModelsLoaded(Message):
    """Top ranked vLLM-capable models were loaded."""

    def __init__(self, mode: str, models: list[ModelCandidate]) -> None:
        super().__init__()
        self.mode = mode
        self.models = models


class VllmModelsFailed(Message):
    """Model discovery failed for a ranking mode."""

    def __init__(self, mode: str, error: str) -> None:
        super().__init__()
        self.mode = mode
        self.error = error


class LlamaCppModelsLoaded(Message):
    """Top ranked llama.cpp-compatible models were loaded."""

    def __init__(self, mode: str, models: list[ModelCandidate]) -> None:
        super().__init__()
        self.mode = mode
        self.models = models


class LlamaCppModelsFailed(Message):
    """llama.cpp model discovery failed for a ranking mode."""

    def __init__(self, mode: str, error: str) -> None:
        super().__init__()
        self.mode = mode
        self.error = error


class LlamaCppQuantsLoaded(Message):
    """Detected GGUF quantizations for a llama.cpp repo."""

    def __init__(
        self,
        repo_id: str,
        revision: str | None,
        quantizations: list[str],
        vram_gb_by_quant: dict[str, float] | None = None,
    ) -> None:
        super().__init__()
        self.repo_id = repo_id
        self.revision = revision
        self.quantizations = quantizations
        self.vram_gb_by_quant = dict(vram_gb_by_quant or {})


class LlamaCppQuantsFailed(Message):
    """GGUF quantization discovery failed for a llama.cpp repo."""

    def __init__(self, repo_id: str, revision: str | None, error: str) -> None:
        super().__init__()
        self.repo_id = repo_id
        self.revision = revision
        self.error = error


class StorageLoaded(Message):
    """Storage snapshot loaded successfully."""

    def __init__(self, snapshot: StorageSnapshot) -> None:
        super().__init__()
        self.snapshot = snapshot


class StorageFailed(Message):
    """Storage listing failed."""

    def __init__(self, error: str) -> None:
        super().__init__()
        self.error = error


# -----------------------------------------------------------------------
# Event dispatcher: protocol event -> Textual message
# -----------------------------------------------------------------------


def _dispatch_event(app_or_widget: object, event: BaseEvent) -> None:
    """Convert a protocol event into a Textual message and post it."""
    poster = getattr(app_or_widget, "post_message", None)
    if poster is None:
        return

    if isinstance(event, LogEvent):
        poster(LogMessage(line=event.line, stream=event.stream))
    elif isinstance(event, StateChangeEvent):
        poster(StateChanged(state=event.current, operation=event.operation, detail=event.detail))
    elif isinstance(event, OperationCompleteEvent):
        poster(
            OperationDone(
                operation=event.operation,
                success=event.success,
                exit_code=event.exit_code,
                detail=event.detail,
                data=event.data,
            )
        )
    elif isinstance(event, ErrorEvent):
        poster(OperationError(message=event.message, recoverable=event.recoverable))



from __future__ import annotations

import unittest

from textual.app import App
from textual.screen import Screen
from textual.widgets import DataTable, Input, OptionList

from llm_launchpad.protocol.enums import BackendType
from llm_launchpad.protocol.models import StorageSnapshot, StoredModelInfo
from llm_launchpad.tui.screens.storage import StorageScreen, _model_label
from llm_launchpad.tui.workers import StorageLoaded


def _sample_snapshot() -> StorageSnapshot:
    return StorageSnapshot(
        llamacpp_models=[
            StoredModelInfo(
                backend=BackendType.LLAMACPP,
                model_id="Qwen/Qwen2.5-Coder-7B-Instruct-GGUF",
                revision="main",
                quant="Q4_K_M",
                size_bytes=1024,
                file_count=1,
                source_volume="huggingface-cache",
                incomplete=True,
            )
        ],
        vllm_models=[
            StoredModelInfo(
                backend=BackendType.VLLM,
                model_id="Qwen/Qwen3-4B-Thinking-2507-FP8",
                revision=None,
                quant=None,
                size_bytes=2048,
                file_count=2,
                source_volume="huggingface-cache",
            )
        ],
    )


class _TestApp(App[None]):
    def __init__(self) -> None:
        super().__init__()
        self.refresh_calls = 0
        self.refresh_force_flags: list[bool] = []
        self.predownload_calls: list[tuple[BackendType, str, str | None, str | None]] = []
        self.delete_calls: list[str] = []

    def begin_storage_refresh(self, receiver: object, force: bool = False) -> None:
        self.refresh_calls += 1
        self.refresh_force_flags.append(force)
        receiver.post_message(StorageLoaded(snapshot=_sample_snapshot()))

    def begin_storage_predownload(  # type: ignore[override]
        self,
        backend: BackendType,
        model_id: str,
        quant: str | None = None,
        revision: str | None = None,
    ) -> None:
        self.predownload_calls.append((backend, model_id, quant, revision))

    def begin_storage_delete(self, model) -> None:  # type: ignore[no-untyped-def]
        self.delete_calls.append(model.model_id)


class _DeferredRefreshApp(_TestApp):
    def __init__(self) -> None:
        super().__init__()
        self._refresh_receiver: object | None = None

    def begin_storage_refresh(self, receiver: object, force: bool = False) -> None:
        self.refresh_calls += 1
        self.refresh_force_flags.append(force)
        self._refresh_receiver = receiver

    def deliver_snapshot(self) -> None:
        assert self._refresh_receiver is not None
        self._refresh_receiver.post_message(StorageLoaded(snapshot=_sample_snapshot()))


class StorageScreenTests(unittest.IsolatedAsyncioTestCase):
    def test_model_label_appends_incomplete_suffix(self) -> None:
        row = StoredModelInfo(
            backend=BackendType.LLAMACPP,
            model_id="Qwen/Qwen2.5-Coder-7B-Instruct-GGUF",
            source_volume="huggingface-cache",
            incomplete=True,
        )
        self.assertEqual(
            _model_label(row),
            "Qwen/Qwen2.5-Coder-7B-Instruct-GGUF (INCOMPLETE)",
        )

    async def test_mount_loads_snapshot_into_table(self) -> None:
        app = _TestApp()
        async with app.run_test() as pilot:
            app.push_screen(StorageScreen())
            await pilot.pause()
            screen = app.screen
            assert isinstance(screen, StorageScreen)
            table = screen.query_one("#storage-table", DataTable)
            self.assertEqual(table.row_count, 2)

    async def test_mount_focuses_backend_filter_menu_for_arrow_navigation(self) -> None:
        app = _TestApp()
        async with app.run_test(size=(100, 35)) as pilot:
            app.push_screen(StorageScreen(initial_backend=BackendType.VLLM))
            await pilot.pause()
            screen = app.screen
            assert isinstance(screen, StorageScreen)
            backend_filter = screen.query_one("#storage-backend-filter", OptionList)
            self.assertTrue(backend_filter.has_focus)

            highlighted = backend_filter.highlighted_option
            self.assertIsNotNone(highlighted)
            assert highlighted is not None
            self.assertEqual(highlighted.id, "filter-vllm")

            await pilot.press("up")
            await pilot.pause()
            highlighted = backend_filter.highlighted_option
            self.assertIsNotNone(highlighted)
            assert highlighted is not None
            self.assertEqual(highlighted.id, "filter-llamacpp")

    async def test_down_from_backend_filter_moves_focus_to_action_menu(self) -> None:
        app = _TestApp()
        async with app.run_test(size=(100, 35)) as pilot:
            app.push_screen(StorageScreen(initial_backend=BackendType.VLLM))
            await pilot.pause()
            screen = app.screen
            assert isinstance(screen, StorageScreen)
            backend_filter = screen.query_one("#storage-backend-filter", OptionList)
            action_list = screen.query_one("#storage-action-list", OptionList)

            backend_filter.focus()
            backend_filter.highlighted = 2
            await pilot.pause()

            await pilot.press("down")
            await pilot.pause()

            self.assertTrue(action_list.has_focus)
            highlighted = action_list.highlighted_option
            self.assertIsNotNone(highlighted)
            assert highlighted is not None
            self.assertEqual(highlighted.id, "refresh")

    async def test_down_from_action_list_last_option_moves_focus_to_table(self) -> None:
        app = _TestApp()
        async with app.run_test(size=(100, 35)) as pilot:
            app.push_screen(StorageScreen())
            await pilot.pause()
            screen = app.screen
            assert isinstance(screen, StorageScreen)
            action_list = screen.query_one("#storage-action-list", OptionList)
            table = screen.query_one("#storage-table", DataTable)

            action_list.focus()
            action_list.highlighted = 2
            await pilot.pause()
            await pilot.press("down")
            await pilot.pause()

            self.assertTrue(table.has_focus)

    async def test_down_inside_table_moves_row_until_last_then_focuses_model_id(self) -> None:
        app = _TestApp()
        async with app.run_test() as pilot:
            app.push_screen(StorageScreen())
            await pilot.pause()
            screen = app.screen
            assert isinstance(screen, StorageScreen)
            table = screen.query_one("#storage-table", DataTable)
            model_id = screen.query_one("#storage-model-id", Input)

            table.focus()
            table.move_cursor(row=0)
            await pilot.pause()
            await pilot.press("down")
            await pilot.pause()
            self.assertTrue(table.has_focus)
            self.assertEqual(table.cursor_row, 1)

            await pilot.press("down")
            await pilot.pause()
            self.assertTrue(model_id.has_focus)

    async def test_up_from_table_first_row_moves_focus_to_action_list(self) -> None:
        app = _TestApp()
        async with app.run_test(size=(100, 35)) as pilot:
            app.push_screen(StorageScreen())
            await pilot.pause()
            screen = app.screen
            assert isinstance(screen, StorageScreen)
            table = screen.query_one("#storage-table", DataTable)
            action_list = screen.query_one("#storage-action-list", OptionList)

            table.focus()
            table.move_cursor(row=0)
            await pilot.pause()
            await pilot.press("up")
            await pilot.pause()

            self.assertTrue(action_list.has_focus)

    async def test_small_viewport_skips_collapsed_option_lists(self) -> None:
        app = _TestApp()
        async with app.run_test(size=(100, 24)) as pilot:
            app.push_screen(StorageScreen())
            await pilot.pause()
            screen = app.screen
            assert isinstance(screen, StorageScreen)
            backend_filter = screen.query_one("#storage-backend-filter", OptionList)
            action_list = screen.query_one("#storage-action-list", OptionList)
            table = screen.query_one("#storage-table", DataTable)

            self.assertEqual(backend_filter.size.height, 0)
            self.assertEqual(action_list.size.height, 0)
            self.assertTrue(table.has_focus)

    async def test_small_viewport_table_boundary_moves_to_model_id(self) -> None:
        app = _TestApp()
        async with app.run_test(size=(100, 24)) as pilot:
            app.push_screen(StorageScreen())
            await pilot.pause()
            screen = app.screen
            assert isinstance(screen, StorageScreen)
            table = screen.query_one("#storage-table", DataTable)
            model_id = screen.query_one("#storage-model-id", Input)

            self.assertTrue(table.has_focus)
            table.move_cursor(row=0)
            await pilot.pause()
            await pilot.press("down")
            await pilot.pause()
            self.assertEqual(table.cursor_row, 1)

            await pilot.press("down")
            await pilot.pause()
            self.assertTrue(model_id.has_focus)

    async def test_initial_storage_load_refocuses_table_in_small_viewport(self) -> None:
        app = _DeferredRefreshApp()
        async with app.run_test(size=(100, 24)) as pilot:
            app.push_screen(StorageScreen())
            await pilot.pause()
            screen = app.screen
            assert isinstance(screen, StorageScreen)
            table = screen.query_one("#storage-table", DataTable)
            model_id = screen.query_one("#storage-model-id", Input)

            model_id.focus()
            await pilot.pause()
            self.assertTrue(model_id.has_focus)

            app.deliver_snapshot()
            await pilot.pause()

            self.assertEqual(table.row_count, 2)
            self.assertTrue(table.has_focus)

    async def test_predownload_uses_form_values(self) -> None:
        app = _TestApp()
        async with app.run_test() as pilot:
            app.push_screen(StorageScreen())
            await pilot.pause()
            screen = app.screen
            assert isinstance(screen, StorageScreen)
            screen.query_one("#storage-model-id", Input).value = "Qwen/Qwen2.5-Coder-7B-Instruct-GGUF"
            screen.query_one("#storage-model-backend", Input).value = "llamacpp"
            screen.query_one("#storage-model-quant", Input).value = "Q4_K_M"
            screen.query_one("#storage-model-revision", Input).value = "main"
            screen.action_predownload_selected()
            self.assertEqual(len(app.predownload_calls), 1)
            backend, model_id, quant, revision = app.predownload_calls[0]
            self.assertEqual(backend, BackendType.LLAMACPP)
            self.assertEqual(model_id, "Qwen/Qwen2.5-Coder-7B-Instruct-GGUF")
            self.assertEqual(quant, "Q4_K_M")
            self.assertEqual(revision, "main")

    async def test_delete_selected_model_uses_selected_row(self) -> None:
        app = _TestApp()
        async with app.run_test() as pilot:
            app.push_screen(StorageScreen())
            await pilot.pause()
            screen = app.screen
            assert isinstance(screen, StorageScreen)
            first_model = screen._snapshot.llamacpp_models[0]
            screen._selected_model = first_model
            screen.action_delete_selected_model()
            self.assertEqual(app.delete_calls, [first_model.model_id])

    async def test_keyboard_navigation_prefills_model_fields(self) -> None:
        app = _TestApp()
        async with app.run_test() as pilot:
            app.push_screen(StorageScreen())
            await pilot.pause()
            screen = app.screen
            assert isinstance(screen, StorageScreen)
            table = screen.query_one("#storage-table", DataTable)
            table.focus()
            await pilot.pause()
            await pilot.press("down")
            await pilot.pause()
            self.assertEqual(
                screen.query_one("#storage-model-id", Input).value,
                "Qwen/Qwen3-4B-Thinking-2507-FP8",
            )
            self.assertEqual(screen.query_one("#storage-model-backend", Input).value, "vllm")

    async def test_resume_triggers_storage_refresh(self) -> None:
        app = _TestApp()
        async with app.run_test() as pilot:
            app.push_screen(StorageScreen())
            await pilot.pause()
            self.assertEqual(app.refresh_calls, 1)
            app.push_screen(Screen())
            await pilot.pause()
            app.pop_screen()
            await pilot.pause()
            self.assertEqual(app.refresh_calls, 2)

    async def test_manual_refresh_uses_force_refresh(self) -> None:
        app = _TestApp()
        async with app.run_test() as pilot:
            app.push_screen(StorageScreen())
            await pilot.pause()
            screen = app.screen
            assert isinstance(screen, StorageScreen)
            screen.action_refresh_storage()
            await pilot.pause()
            self.assertEqual(app.refresh_force_flags, [False, True])


if __name__ == "__main__":
    unittest.main()



"""Instance/app naming helpers for launchpad deployments."""

from __future__ import annotations

import re
import unicodedata
from typing import Optional

from coolname import generate_slug

from ..protocol.enums import BackendType


_LEGACY_APP_NAMES = {
    BackendType.LLAMACPP: "llamacpp-server",
    BackendType.VLLM: "vllm-server",
}

_BACKEND_PREFIX = {
    BackendType.LLAMACPP: "llamacpp",
    BackendType.VLLM: "vllm",
}


def legacy_app_name(backend: BackendType) -> str:
    """Return the historical single-instance app name for *backend*."""
    return _LEGACY_APP_NAMES[backend]


def slugify_instance_name(raw: str, default: str = "default") -> str:
    """Convert an arbitrary identifier into a safe, deterministic slug."""
    normalized = unicodedata.normalize("NFKD", raw or "")
    ascii_text = normalized.encode("ascii", "ignore").decode("ascii").lower()
    ascii_text = ascii_text.replace("/", "-").replace("_", "-")
    slug = re.sub(r"[^a-z0-9-]+", "-", ascii_text)
    slug = re.sub(r"-{2,}", "-", slug).strip("-")
    return slug or default


def auto_instance_name_for_backend(backend: BackendType, model_hint: Optional[str]) -> str:
    """Build a default instance name based on backend model context."""
    if backend == BackendType.VLLM:
        return slugify_instance_name(model_hint or "default")
    # llama.cpp can use preset name or repo id as a hint; fallback still valid.
    return slugify_instance_name(model_hint or "default")


def default_served_model_name(model_name: Optional[str], default: str = "llm") -> str:
    """Return the default served model alias for a vLLM model id."""
    candidate = (model_name or "").strip()
    if not candidate:
        return default
    tail = candidate.rsplit("/", 1)[-1].strip()
    return tail or default


def default_llamacpp_served_model_name(
    repo_id: Optional[str],
    quant: Optional[str] = None,
    default: str = "default",
) -> str:
    """Return a friendly llama.cpp OpenAI model id derived from repo + quant."""
    alias = default_served_model_name(repo_id, default=default)
    quant_text = (quant or "").strip()
    if not quant_text:
        return alias
    if quant_text.casefold() in alias.casefold():
        return alias
    return f"{alias}-{quant_text}"


def build_app_name(backend: BackendType, instance_name: Optional[str]) -> str:
    """Compose a launchpad app name from backend + instance name."""
    if not instance_name:
        return legacy_app_name(backend)
    return f"{_BACKEND_PREFIX[backend]}-{slugify_instance_name(instance_name)}"


def infer_backend_from_app_name(app_name: str) -> Optional[BackendType]:
    """Infer backend type from legacy or prefixed app names."""
    if app_name == _LEGACY_APP_NAMES[BackendType.VLLM] or app_name.startswith("vllm-"):
        return BackendType.VLLM
    if app_name == _LEGACY_APP_NAMES[BackendType.LLAMACPP] or app_name.startswith("llamacpp-"):
        return BackendType.LLAMACPP
    return None


def infer_instance_from_app_name(app_name: str, backend: Optional[BackendType]) -> Optional[str]:
    """Infer instance name from app name when possible."""
    if backend is None:
        return None
    legacy = legacy_app_name(backend)
    if app_name == legacy:
        return "default"
    prefix = _BACKEND_PREFIX[backend] + "-"
    if app_name.startswith(prefix):
        return app_name[len(prefix) :] or "default"
    return None


def random_function_slug() -> str:
    """Generate a random, URL-safe two-word slug for Modal function names."""
    return slugify_instance_name(generate_slug(2))


def modal_function_name(base_name: str, function_slug: Optional[str]) -> str:
    """Return a Modal function name with optional deployment slug suffix."""
    slug = slugify_instance_name(function_slug or "", default="")
    if not slug:
        return base_name
    return f"{base_name}-{slug}"



"""Shared helpers for deployment GPU type/count handling."""

from __future__ import annotations

from collections.abc import Sequence

from ..core.modal_gpu import ModalGpuSpec

DEFAULT_GPU_TYPE = "A100-80GB"
DEFAULT_GPU_COUNT = 1


def normalize_gpu_type(value: str) -> str:
    return value.strip().upper()


def parse_gpu_count(value: str, default: int = DEFAULT_GPU_COUNT) -> int:
    raw = value.strip()
    if not raw:
        return default
    try:
        count = int(raw)
    except ValueError:
        return default
    return count if count > 0 else default


def build_gpu_type_options(gpu_types: Sequence[str | ModalGpuSpec]) -> list[tuple[str, str]]:
    """Build `(label, value)` pairs for GPU type dropdown options."""
    options: list[tuple[str, str]] = []
    seen: set[str] = set()
    for value in gpu_types:
        if isinstance(value, ModalGpuSpec):
            token = normalize_gpu_type(value.value)
            label = _format_gpu_type_label(token, value.price_per_hour_usd)
        else:
            token = normalize_gpu_type(value)
            label = token
        if not token or token in seen:
            continue
        seen.add(token)
        options.append((label, token))
    return options


def _format_gpu_type_label(value: str, price_per_hour_usd: float | None) -> str:
    if price_per_hour_usd is None:
        return value
    return f"{value} (${price_per_hour_usd:.2f}/hr)"



"""Shared screen behavior for native-first copy-to-clipboard support."""

from __future__ import annotations

from typing import Any

from rich.errors import MarkupError
from rich.markup import render as render_markup
from textual.binding import Binding
from textual.selection import Selection
from textual.screen import Screen
from textual.widgets import DataTable, Input, OptionList, Select, Static


class CopyEnabledScreen(Screen):
    """Screen base class that provides consistent copy behavior."""

    BINDINGS = [
        Binding("y", "copy_text", "Copy", key_display="y", show=False),
        Binding(
            "ctrl+shift+c,super+c,meta+c,cmd+c,command+c",
            "copy_text_to_clipboard",
            "Copy",
            show=False,
        ),
    ]

    def __init__(self, *args: object, **kwargs: object) -> None:
        super().__init__(*args, **kwargs)
        self._selection_sync_scheduled = False
        self._last_synced_selection: str | None = None

    def action_copy_text(self) -> None:
        """Copy selected or focused text to the clipboard."""
        normalized = self._copyable_text()
        if normalized:
            self.app.copy_to_clipboard(normalized)
            return
        self.notify("Nothing to copy", timeout=2)

    def action_copy_text_to_clipboard(self) -> None:
        """Copy selected text directly to the clipboard when the terminal supports it."""
        normalized = self._copyable_text()
        if normalized:
            self.app.copy_to_clipboard(normalized)
            return
        self.notify("Nothing to copy", timeout=2)

    async def _watch_selections(
        self,
        old_selections: dict[Static, Selection],
        selections: dict[Static, Selection],
    ) -> None:
        await super()._watch_selections(old_selections, selections)
        if not selections:
            self._last_synced_selection = None
            return
        if self._selection_sync_scheduled:
            return
        self._selection_sync_scheduled = True
        self.call_after_refresh(self._sync_selection_to_clipboard)

    def _copyable_text(self) -> str | None:
        return self._normalize_text(
            self._selected_text_for_copy() or self._focused_text_fallback()
        )

    def _sync_selection_to_clipboard(self) -> None:
        self._selection_sync_scheduled = False
        selection = self._normalize_text(self._selected_text_for_copy())
        if not selection:
            self._last_synced_selection = None
            return
        if selection == self._last_synced_selection:
            return
        self.app.copy_to_clipboard(selection)
        self._last_synced_selection = selection

    def _selected_text_for_copy(self) -> str | None:
        return self.get_selected_text()

    def _focused_text_fallback(self) -> str | None:
        focused = self.focused
        if isinstance(focused, OptionList):
            return self._copy_from_option_list(focused)
        if isinstance(focused, DataTable):
            return self._copy_from_data_table(focused)
        if isinstance(focused, Input):
            return focused.value
        if isinstance(focused, Select):
            selected = focused.selection
            return str(selected) if selected is not None else None
        if isinstance(focused, Static):
            return self._to_plain_text(focused.render())
        return None

    def _copy_from_option_list(self, widget: OptionList) -> str | None:
        highlighted = widget.highlighted
        if highlighted is None:
            return None
        try:
            option = widget.get_option_at_index(highlighted)
        except Exception:
            return None
        return self._to_plain_text(option.prompt)

    def _copy_from_data_table(self, widget: DataTable) -> str | None:
        if widget.row_count <= 0:
            return None
        try:
            row_values = widget.get_row_at(widget.cursor_row)
        except Exception:
            return None
        return "\t".join(self._to_plain_text(value) for value in row_values)

    def _normalize_text(self, text: str | None) -> str | None:
        if text is None:
            return None
        normalized = text.rstrip("\n")
        return normalized if normalized else None

    def _to_plain_text(self, value: Any) -> str:
        if value is None:
            return ""
        if isinstance(value, str):
            return self._strip_markup(value)
        return str(value)

    def _strip_markup(self, text: str) -> str:
        try:
            return render_markup(text).plain
        except MarkupError:
            return text



from __future__ import annotations

import unittest

from llm_launchpad.core.backend import ModalBackend
from llm_launchpad.protocol.enums import BackendType
from llm_launchpad.protocol.models import DeploymentConfig, LaunchpadSettings


class BackendEnvTests(unittest.TestCase):
    def test_vllm_reasoning_and_tool_fields_are_forwarded(self) -> None:
        config = DeploymentConfig(
            backend=BackendType.VLLM,
            model_name="Qwen/Qwen3-8B",
            trust_remote_code=True,
            reasoning_parser="qwen3",
            tool_call_parser="qwen3_xml",
            default_chat_template_kwargs='{"enable_thinking": false}',
        )

        env = ModalBackend.env_for_backend(config)
        self.assertEqual(env["MODEL_NAME"], "Qwen/Qwen3-8B")
        self.assertEqual(env["TRUST_REMOTE_CODE"], "true")
        self.assertEqual(env["REASONING_PARSER"], "qwen3")
        self.assertEqual(env["TOOL_CALL_PARSER"], "qwen3_xml")
        self.assertEqual(env["DEFAULT_CHAT_TEMPLATE_KWARGS"], '{"enable_thinking": false}')

    def test_vllm_reasoning_and_tool_fields_are_omitted_when_empty(self) -> None:
        config = DeploymentConfig(
            backend=BackendType.VLLM,
            model_name="Qwen/Qwen3-8B",
            trust_remote_code=None,
            reasoning_parser=None,
            tool_call_parser="",
            default_chat_template_kwargs="",
        )

        env = ModalBackend.env_for_backend(config)
        self.assertNotIn("TRUST_REMOTE_CODE", env)
        self.assertNotIn("REASONING_PARSER", env)
        self.assertNotIn("TOOL_CALL_PARSER", env)
        self.assertNotIn("DEFAULT_CHAT_TEMPLATE_KWARGS", env)

    def test_build_full_env_uses_deployment_gpu_config(self) -> None:
        settings = LaunchpadSettings(scaledown_window=900)
        config = DeploymentConfig(
            backend=BackendType.LLAMACPP,
            gpu_type="H100",
            gpu_count=2,
        )
        env = ModalBackend.build_full_env(settings, config)
        self.assertEqual(env["GPU_CONFIG"], "H100:2")
        self.assertEqual(env["SCALEDOWN_WINDOW"], "900")

    def test_build_full_env_ignores_blank_gpu_type(self) -> None:
        settings = LaunchpadSettings()
        config = DeploymentConfig(
            backend=BackendType.LLAMACPP,
            gpu_type="   ",
            gpu_count=2,
        )
        env = ModalBackend.build_full_env(settings, config)
        self.assertNotIn("GPU_CONFIG", env)

    def test_vllm_n_gpu_is_independent_from_deployment_gpu_count(self) -> None:
        settings = LaunchpadSettings()
        config = DeploymentConfig(
            backend=BackendType.VLLM,
            gpu_type="A100-80GB",
            gpu_count=2,
            n_gpu=4,
        )
        env = ModalBackend.build_full_env(settings, config)
        self.assertEqual(env["GPU_CONFIG"], "A100-80GB:2")
        self.assertEqual(env["N_GPU"], "4")

    def test_env_for_backend_includes_function_slug(self) -> None:
        config = DeploymentConfig(
            backend=BackendType.LLAMACPP,
            app_name="llamacpp-prod",
            function_slug="alpha-bravo",
        )
        env = ModalBackend.env_for_backend(config)
        self.assertEqual(env["MODAL_APP_NAME"], "llamacpp-prod")
        self.assertEqual(env["MODAL_FUNCTION_SLUG"], "alpha-bravo")


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

import os
import subprocess
import tempfile
import unittest
from unittest.mock import patch

from llm_launchpad.core.backend import ModalBackend, _extract_modal_app_rows
from llm_launchpad.protocol.enums import BackendType
from llm_launchpad.protocol.models import DeploymentConfig


class BackendParsingAndCommandTests(unittest.TestCase):
    def test_modal_cli_path_prefers_active_env_scripts_dir(self) -> None:
        with tempfile.TemporaryDirectory() as tmp:
            bin_dir = os.path.join(tmp, "bin")
            os.makedirs(bin_dir, exist_ok=True)
            modal_path = os.path.join(bin_dir, "modal")
            with open(modal_path, "w", encoding="utf-8") as fh:
                fh.write("#!/bin/sh\n")
            with (
                patch("llm_launchpad.core.backend.sys.prefix", tmp),
                patch("llm_launchpad.core.backend.shutil.which", return_value=None),
            ):
                self.assertEqual(ModalBackend.modal_cli_path(), modal_path)

    @patch("llm_launchpad.core.backend.subprocess.run")
    def test_list_apps_returns_empty_list_for_empty_json_payload(self, mock_run) -> None:  # type: ignore[no-untyped-def]
        mock_run.return_value.returncode = 0
        mock_run.return_value.stdout = "[]"
        self.assertEqual(ModalBackend.list_apps(), [])

    @patch("llm_launchpad.core.backend.subprocess.run")
    def test_list_apps_returns_none_on_timeout(self, mock_run) -> None:  # type: ignore[no-untyped-def]
        mock_run.side_effect = subprocess.TimeoutExpired(cmd=["modal", "app", "list"], timeout=8)
        self.assertIsNone(ModalBackend.list_apps())

    @patch("llm_launchpad.core.backend.subprocess.run")
    def test_list_volume_returns_none_on_timeout(self, mock_run) -> None:  # type: ignore[no-untyped-def]
        mock_run.side_effect = subprocess.TimeoutExpired(cmd=["modal", "volume", "ls"], timeout=8)
        self.assertIsNone(ModalBackend.list_volume("huggingface-cache", "/hub"))

    @patch("llm_launchpad.core.backend.subprocess.run")
    def test_billing_report_json_returns_payload(self, mock_run) -> None:  # type: ignore[no-untyped-def]
        mock_run.return_value.returncode = 0
        mock_run.return_value.stdout = '{"summary":{"total_usd":3.25}}'
        payload, error = ModalBackend.billing_report_json()
        self.assertIsInstance(payload, dict)
        self.assertIsNone(error)
        assert isinstance(payload, dict)
        self.assertEqual(payload["summary"]["total_usd"], 3.25)
        called_command = mock_run.call_args.args[0]
        self.assertEqual(
            called_command[-5:],
            ["billing", "report", "--for", "this month", "--json"],
        )

    @patch("llm_launchpad.core.backend.subprocess.run")
    def test_billing_report_json_returns_none_on_timeout(self, mock_run) -> None:  # type: ignore[no-untyped-def]
        mock_run.side_effect = subprocess.TimeoutExpired(
            cmd=["modal", "billing", "report", "--for", "this month", "--json"],
            timeout=8,
        )
        payload, error = ModalBackend.billing_report_json()
        self.assertIsNone(payload)
        self.assertIn("Timed out", error or "")

    @patch("llm_launchpad.core.backend.subprocess.run")
    def test_billing_report_json_falls_back_to_workspace_command(self, mock_run) -> None:  # type: ignore[no-untyped-def]
        first = subprocess.CompletedProcess(
            args=["modal", "billing", "report", "--for", "this month", "--json"],
            returncode=2,
            stdout="",
            stderr="No such command 'billing'.",
        )
        second = subprocess.CompletedProcess(
            args=["modal", "workspace", "billing", "report", "--for", "this month", "--json"],
            returncode=0,
            stdout='{"summary":{"total_usd":7.0}}',
            stderr="",
        )
        mock_run.side_effect = [first, second]

        payload, error = ModalBackend.billing_report_json()
        self.assertIsNone(error)
        assert isinstance(payload, dict)
        self.assertEqual(payload["summary"]["total_usd"], 7.0)

    @patch("llm_launchpad.core.backend.subprocess.run")
    def test_billing_report_json_returns_upgrade_hint_when_billing_command_missing(self, mock_run) -> None:  # type: ignore[no-untyped-def]
        first = subprocess.CompletedProcess(
            args=["modal", "billing", "report", "--for", "this month", "--json"],
            returncode=2,
            stdout="",
            stderr="No such command 'billing'.",
        )
        second = subprocess.CompletedProcess(
            args=["modal", "workspace", "billing", "report", "--for", "this month", "--json"],
            returncode=2,
            stdout="Usage: modal [OPTIONS] COMMAND [ARGS]...",
            stderr="",
        )
        mock_run.side_effect = [first, second]

        payload, error = ModalBackend.billing_report_json()
        self.assertIsNone(payload)
        self.assertIn("Upgrade Modal CLI", error or "")

    def test_extract_modal_app_rows_accepts_apps_and_data_wrappers(self) -> None:
        wrapped_apps = {
            "apps": [
                {"app_name": "vllm-qwen", "id": "ap-1", "status": "running"},
            ]
        }
        wrapped_data = {
            "data": [
                {"label": "llamacpp-phi", "appId": "ap-2", "phase": "deployed"},
            ]
        }

        rows_apps = _extract_modal_app_rows(wrapped_apps)
        rows_data = _extract_modal_app_rows(wrapped_data)

        self.assertEqual(rows_apps[0].name, "vllm-qwen")
        self.assertEqual(rows_apps[0].app_id, "ap-1")
        self.assertEqual(rows_apps[0].state, "running")
        self.assertEqual(rows_apps[0].backend, BackendType.VLLM)
        self.assertEqual(rows_apps[0].instance_name, "qwen")

        self.assertEqual(rows_data[0].name, "llamacpp-phi")
        self.assertEqual(rows_data[0].app_id, "ap-2")
        self.assertEqual(rows_data[0].state, "deployed")
        self.assertEqual(rows_data[0].backend, BackendType.LLAMACPP)
        self.assertEqual(rows_data[0].instance_name, "phi")

    def test_extract_modal_app_rows_skips_non_dict_and_non_list_payload(self) -> None:
        self.assertEqual(_extract_modal_app_rows("not-a-list"), [])
        rows = _extract_modal_app_rows([None, 1, {"name": "vllm-qwen", "state": "running"}])
        self.assertEqual(len(rows), 1)
        self.assertEqual(rows[0].name, "vllm-qwen")

    def test_extract_modal_app_rows_extracts_web_url_and_model_metadata(self) -> None:
        payload = [
            {
                "name": "vllm-qwen",
                "state": "running",
                "details": {
                    "webUrl": "https://alice--vllm-qwen-serve.modal.run",
                    "env": {
                        "MODEL_NAME": "Qwen/Qwen3-4B-Thinking-2507-FP8",
                        "SERVED_MODEL_NAME": "Qwen3-4B",
                    },
                },
            },
            {
                "name": "llamacpp-phi",
                "state": "deployed",
                "metadata": {
                    "config": {
                        "repo_id": "unsloth/Phi-3-GGUF",
                        "quant": "Q4_K_M",
                    }
                },
                "url": "https://alice--llamacpp-phi-serve-sly-otter.modal.run",
            },
        ]

        rows = _extract_modal_app_rows(payload)
        self.assertEqual(rows[0].web_url, "https://alice--vllm-qwen-serve.modal.run")
        self.assertEqual(rows[0].model_name, "Qwen/Qwen3-4B-Thinking-2507-FP8")
        self.assertEqual(rows[0].served_model_name, "Qwen3-4B")
        self.assertEqual(rows[1].web_url, "https://alice--llamacpp-phi-serve-sly-otter.modal.run")
        self.assertEqual(rows[1].repo_id, "unsloth/Phi-3-GGUF")
        self.assertEqual(rows[1].quant, "Q4_K_M")

    def test_extract_modal_app_rows_infers_backend_from_metadata_for_custom_names(self) -> None:
        payload = [
            {
                "name": "prod-coder",
                "state": "running",
                "details": {
                    "webUrl": "https://alice--prod-coder-serve.modal.run",
                    "env": {
                        "MODEL_NAME": "Qwen/Qwen3-4B-Thinking-2507-FP8",
                        "SERVED_MODEL_NAME": "Qwen3-4B",
                    },
                },
            },
            {
                "name": "edge-gguf",
                "state": "deployed",
                "metadata": {
                    "config": {
                        "repo_id": "unsloth/Phi-3-GGUF",
                        "quant": "Q4_K_M",
                    }
                },
            },
        ]

        rows = _extract_modal_app_rows(payload)

        self.assertEqual(rows[0].backend, BackendType.VLLM)
        self.assertIsNone(rows[0].instance_name)
        self.assertEqual(rows[1].backend, BackendType.LLAMACPP)
        self.assertIsNone(rows[1].instance_name)

    def test_build_run_command_llamacpp_includes_full_option_set(self) -> None:
        config = DeploymentConfig(
            backend=BackendType.LLAMACPP,
            preset="qwen3-coder-30b",
            repo_id="unsloth/Qwen3-Coder-30B-A3B-Instruct-1M-GGUF",
            quant="Q4_K_M",
            revision="main",
            preload=False,
            do_deploy=True,
            server_args="--ctx-size 65536",
            host="0.0.0.0",
            port=8080,
            n_gpu_layers=100,
        )
        cmd = ModalBackend.build_run_command(config)
        self.assertIn("--preset", cmd)
        self.assertIn("--repo-id", cmd)
        self.assertIn("--quant", cmd)
        self.assertIn("--revision", cmd)
        self.assertIn("--no-preload", cmd)
        self.assertIn("--deploy", cmd)
        self.assertIn("--server-args=--ctx-size 65536", cmd)
        self.assertIn("--host", cmd)
        self.assertIn("--port", cmd)
        self.assertIn("--n-gpu-layers", cmd)

    def test_env_for_backend_non_vllm_only_sets_app_name(self) -> None:
        config = DeploymentConfig(
            backend=BackendType.LLAMACPP,
            app_name="llamacpp-prod",
            model_name="should-not-be-used",
            served_model_name="ignored",
        )
        env = ModalBackend.env_for_backend(config)
        self.assertEqual(env, {"MODAL_APP_NAME": "llamacpp-prod"})

    def test_env_for_backend_llamacpp_forwards_image_no_cache_toggle(self) -> None:
        config = DeploymentConfig(
            backend=BackendType.LLAMACPP,
            app_name="llamacpp-prod",
            llamacpp_image_no_cache=True,
        )
        env = ModalBackend.env_for_backend(config)
        self.assertEqual(
            env,
            {
                "MODAL_APP_NAME": "llamacpp-prod",
                "LLAMA_CPP_IMAGE_NO_CACHE": "true",
            },
        )

    def test_test_curl_command_vllm_prefers_provided_served_model_name(self) -> None:
        cmd = ModalBackend.test_curl_command(
            BackendType.VLLM,
            "https://example.modal.run",
            served_model_name="Qwen3-0.6B",
        )
        self.assertIn('"model":"Qwen3-0.6B"', cmd)

    def test_test_curl_command_llamacpp_is_copy_paste_ready(self) -> None:
        cmd = ModalBackend.test_curl_command(
            BackendType.LLAMACPP,
            "https://example.modal.run/",
        )
        self.assertTrue(
            cmd.startswith("curl -s -X POST https://example.modal.run/v1/completions ")
        )
        self.assertIn('"prompt":"Say hello in one short sentence."', cmd)

    def test_test_curl_command_llamacpp_uses_served_model_name_when_provided(self) -> None:
        cmd = ModalBackend.test_curl_command(
            BackendType.LLAMACPP,
            "https://example.modal.run/",
            served_model_name="Nanbeige4.1-3B-Q4_K_M-GGUF",
        )
        self.assertIn('"model":"Nanbeige4.1-3B-Q4_K_M-GGUF"', cmd)


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

import types
import unittest
from pathlib import Path
from tempfile import NamedTemporaryFile
from types import SimpleNamespace
from unittest.mock import patch

from llm_launchpad.core import hf_models
from llm_launchpad.core.hf_models import ModelCandidate


class HFModelsTests(unittest.TestCase):
    def setUp(self) -> None:
        hf_models._CACHE.clear()
        hf_models._GGUF_QUANT_METADATA_CACHE.clear()
        hf_models._VLLM_MEMORY_CACHE.clear()
        hf_models._HF_JSON_FILE_CACHE.clear()

    def test_normalize_keeps_text_generation_pipeline(self) -> None:
        row = SimpleNamespace(
            id="Qwen/Qwen2.5-7B-Instruct",
            downloads=123,
            likes=45,
            pipeline_tag="text-generation",
            tags=["transformers"],
        )
        candidate = hf_models._normalize_candidate(row)
        self.assertIsNotNone(candidate)
        assert candidate is not None
        self.assertEqual(candidate.repo_id, "Qwen/Qwen2.5-7B-Instruct")
        self.assertEqual(candidate.downloads, 123)
        self.assertEqual(candidate.likes, 45)

    def test_normalize_keeps_text_generation_tag(self) -> None:
        row = SimpleNamespace(
            modelId="meta-llama/Llama-3.1-8B-Instruct",
            downloads="99",
            likes="7",
            pipeline_tag=None,
            tags=["text-generation", "transformers"],
        )
        candidate = hf_models._normalize_candidate(row)
        self.assertIsNotNone(candidate)
        assert candidate is not None
        self.assertEqual(candidate.repo_id, "meta-llama/Llama-3.1-8B-Instruct")
        self.assertEqual(candidate.downloads, 99)
        self.assertEqual(candidate.likes, 7)

    def test_normalize_filters_non_text_generation(self) -> None:
        row = SimpleNamespace(
            id="some/image-model",
            downloads=10,
            likes=1,
            pipeline_tag="image-classification",
            tags=["image-classification"],
        )
        self.assertIsNone(hf_models._normalize_candidate(row))

    def test_normalize_filters_llamacpp_without_gguf_tag(self) -> None:
        row = SimpleNamespace(
            id="Qwen/Qwen3-4B-Instruct",
            downloads=10,
            likes=1,
            pipeline_tag="text-generation",
            tags=["text-generation"],
        )
        self.assertIsNone(hf_models._normalize_candidate(row, target="llamacpp"))

    def test_list_vllm_candidates_uses_ttl_cache(self) -> None:
        expected = [ModelCandidate(repo_id="Qwen/Qwen2.5-7B-Instruct")]
        with patch("llm_launchpad.core.hf_models._fetch_candidates", return_value=expected) as fetch:
            first = hf_models.list_vllm_candidates(mode="downloads", limit=10)
            second = hf_models.list_vllm_candidates(mode="downloads", limit=10)
        self.assertEqual(first, expected)
        self.assertEqual(second, expected)
        self.assertEqual(fetch.call_count, 1)

    def test_list_llamacpp_candidates_uses_ttl_cache(self) -> None:
        expected = [ModelCandidate(repo_id="Qwen/Qwen3-Coder-Next-GGUF")]
        with patch("llm_launchpad.core.hf_models._fetch_candidates", return_value=expected) as fetch:
            first = hf_models.list_llamacpp_candidates(mode="trending", limit=10)
            second = hf_models.list_llamacpp_candidates(mode="trending", limit=10)
        self.assertEqual(first, expected)
        self.assertEqual(second, expected)
        self.assertEqual(fetch.call_count, 1)

    def test_fetch_candidates_applies_sort_and_limit(self) -> None:
        calls: list[tuple[object, str, int]] = []
        rows = [
            SimpleNamespace(
                id="internlm/Intern-S1-Pro",
                downloads=10,
                likes=5,
                pipeline_tag="image-text-to-text",
                tags=["image-text-to-text"],
            ),
            SimpleNamespace(
                id="Qwen/Qwen3-Coder-Next",
                downloads=20,
                likes=9,
                pipeline_tag="text-generation",
                tags=["text-generation"],
            ),
            SimpleNamespace(
                id="zai-org/GLM-5",
                downloads=30,
                likes=12,
                pipeline_tag="text-generation",
                tags=["text-generation"],
            ),
        ]

        class FakeApi:
            def list_models(self, *, filter: object, sort: str, limit: int, full: bool):
                self._ = (filter, full)
                calls.append((filter, sort, limit))
                return rows

        fake_module = types.SimpleNamespace(HfApi=FakeApi)
        with patch.dict("sys.modules", {"huggingface_hub": fake_module}):
            result = hf_models._fetch_candidates(mode="trending", limit=2, target="vllm")
        self.assertEqual(calls, [("text-generation", "trending_score", 6)])
        self.assertEqual([m.repo_id for m in result], ["Qwen/Qwen3-Coder-Next", "zai-org/GLM-5"])

    def test_fetch_llamacpp_candidates_uses_gguf_filter(self) -> None:
        calls: list[tuple[object, str, int]] = []
        rows = [
            SimpleNamespace(
                id="Qwen/Qwen3-Coder-Next-GGUF",
                downloads=20,
                likes=9,
                pipeline_tag="text-generation",
                tags=["text-generation", "gguf"],
                siblings=[
                    SimpleNamespace(rfilename="Qwen3-Coder-Next-Q4_K_M.gguf"),
                    SimpleNamespace(rfilename="Qwen3-Coder-Next-Q8_0.gguf"),
                ],
            ),
            SimpleNamespace(
                id="Qwen/Qwen3-4B",
                downloads=30,
                likes=12,
                pipeline_tag="text-generation",
                tags=["text-generation"],
            ),
        ]

        class FakeApi:
            def list_models(self, *, filter: object, sort: str, limit: int, full: bool):
                self._ = (filter, full)
                calls.append((filter, sort, limit))
                return rows

        fake_module = types.SimpleNamespace(HfApi=FakeApi)
        with patch.dict("sys.modules", {"huggingface_hub": fake_module}):
            result = hf_models._fetch_candidates(mode="downloads", limit=2, target="llamacpp")
        self.assertEqual(calls, [(["text-generation", "gguf"], "downloads", 6)])
        self.assertEqual([m.repo_id for m in result], ["Qwen/Qwen3-Coder-Next-GGUF"])
        self.assertEqual(result[0].quantizations, ("Q4_K_M", "Q8_0"))

    def test_extract_gguf_quantizations_parses_and_sorts(self) -> None:
        siblings = [
            SimpleNamespace(rfilename="BF16/model-BF16-00001-of-00002.gguf"),
            SimpleNamespace(rfilename="Q5_K_M/model-Q5_K_M.gguf"),
            SimpleNamespace(rfilename="Q4_K_M/model-Q4_K_M.gguf"),
            SimpleNamespace(rfilename="Q8_0/model-Q8_0.gguf"),
            SimpleNamespace(rfilename="README.md"),
        ]
        self.assertEqual(hf_models._extract_gguf_quantizations(siblings), ["Q4_K_M", "Q5_K_M", "Q8_0"])

    def test_fetch_gguf_quant_metadata_uses_cache(self) -> None:
        calls: list[tuple[str, str | None]] = []

        class FakeApi:
            def model_info(self, *, repo_id: str, revision: str | None, expand: list[str]):
                self._ = expand
                calls.append((repo_id, revision))
                return SimpleNamespace(
                    siblings=[
                        SimpleNamespace(rfilename="Q4_K_M/model-Q4_K_M.gguf"),
                        SimpleNamespace(rfilename="Q6_K/model-Q6_K.gguf"),
                    ],
                    gguf={
                        "compatibility": [
                            {"quantization": "Q4_K_M", "memory": "4.66 GB"},
                            {"quantization": "Q6_K", "memory": "7.2 GB"},
                            {"quantization": "Q5_K_M", "memory": "6.0 GB"},
                        ]
                    },
                )

        fake_module = types.SimpleNamespace(HfApi=FakeApi)
        with (
            patch.dict("sys.modules", {"huggingface_hub": fake_module}),
            patch("llm_launchpad.core.hf_models._fetch_gguf_quantization_data_from_model_page", return_value=None),
        ):
            first = hf_models.fetch_gguf_quant_metadata("Qwen/Qwen3-Coder-Next-GGUF")
            second = hf_models.fetch_gguf_quant_metadata("Qwen/Qwen3-Coder-Next-GGUF")

        self.assertEqual(first.quantizations, ["Q4_K_M", "Q6_K"])
        self.assertEqual(second.quantizations, ["Q4_K_M", "Q6_K"])
        self.assertAlmostEqual(first.vram_gb_by_quant["Q4_K_M"], 4.66, places=2)
        self.assertAlmostEqual(first.vram_gb_by_quant["Q6_K"], 7.2, places=2)
        self.assertNotIn("Q5_K_M", first.vram_gb_by_quant)
        self.assertEqual(calls, [("Qwen/Qwen3-Coder-Next-GGUF", None)])

    def test_parse_memory_to_gb_supports_gb_tb_and_bytes(self) -> None:
        self.assertAlmostEqual(hf_models._parse_memory_to_gb("4.66 GB") or 0, 4.66, places=2)
        self.assertAlmostEqual(hf_models._parse_memory_to_gb("1.5 TB") or 0, 1500.0, places=1)
        self.assertAlmostEqual(hf_models._parse_memory_to_gb(5 * 1024**3) or 0, 5.369, places=3)

    def test_extract_gguf_vram_by_quant_keeps_max_duplicate(self) -> None:
        payload = {
            "hardware_compatibility": [
                {"quant": "Q4_K_M", "vram": "4.6 GB"},
                {"quant": "Q4_K_M", "vram": "4.9 GB"},
                {"quant": "Q6_K", "vram": "7.3 GB"},
            ]
        }
        parsed = hf_models._extract_gguf_vram_by_quant(payload)
        self.assertAlmostEqual(parsed["Q4_K_M"], 4.9, places=1)
        self.assertAlmostEqual(parsed["Q6_K"], 7.3, places=1)

    def test_extract_quantization_data_from_model_page_html(self) -> None:
        html = (
            '<div data-target="ModelTensorsParams" '
            'data-props="{&quot;ggufQuantizationData&quot;:{&quot;variantsByQuantizationLevels&quot;:'
            '{&quot;4&quot;:[{&quot;label&quot;:&quot;Q4_1&quot;,&quot;size&quot;:472704107744}]}}}"></div>'
        )
        parsed = hf_models._extract_gguf_quantization_data_from_model_page_html(html)
        assert isinstance(parsed, dict)
        levels = parsed.get("variantsByQuantizationLevels")
        assert isinstance(levels, dict)
        row = levels.get("4", [])[0]
        self.assertEqual(row["label"], "Q4_1")
        self.assertEqual(row["size"], 472704107744)

    def test_extract_quantizations_and_vram_from_quantization_data_keeps_hf_labels(self) -> None:
        quant_data = {
            "variantsByQuantizationLevels": {
                "1": [{"label": "UD-IQ1_S", "size": 203699011168}],
                "4": [
                    {"label": "MXFP4_MOE", "size": 410557924192},
                    {"label": "Q4_1", "size": 472704107744},
                ],
                "16": [{"label": "BF16", "size": 1510000000000}],
            }
        }
        quantizations, vram = hf_models._extract_quantizations_and_vram_from_quantization_data(quant_data)
        self.assertEqual(quantizations, ["UD-IQ1_S", "MXFP4_MOE", "Q4_1", "BF16"])
        self.assertAlmostEqual(vram["UD-IQ1_S"], 203.699, places=2)
        self.assertAlmostEqual(vram["MXFP4_MOE"], 410.558, places=2)
        self.assertAlmostEqual(vram["Q4_1"], 472.704, places=2)
        self.assertAlmostEqual(vram["BF16"], 1510.0, places=1)

    def test_fetch_quant_metadata_falls_back_to_model_page_quantization_data(self) -> None:
        class FakeApi:
            def model_info(self, *, repo_id: str, revision: str | None, expand: list[str]):
                self._ = (repo_id, revision, expand)
                return SimpleNamespace(
                    siblings=[
                        SimpleNamespace(rfilename="Q4_1/model-Q4_1.gguf"),
                        SimpleNamespace(rfilename="Q8_0/model-Q8_0.gguf"),
                    ],
                    gguf={},
                )

        class FakeResponse:
            status_code = 200
            text = (
                '<div data-target="ModelTensorsParams" '
                'data-props="{&quot;ggufQuantizationData&quot;:{&quot;variantsByQuantizationLevels&quot;:'
                "{&quot;4&quot;:[{&quot;label&quot;:&quot;Q4_1&quot;,&quot;size&quot;:472704107744}],"
                '&quot;8&quot;:[{&quot;label&quot;:&quot;Q8_0&quot;,&quot;size&quot;:801344964608}]'
                "}}}\"></div>"
            )

        def fake_get(url: str, timeout: float, headers: dict[str, str]):
            self.assertIn("huggingface.co", url)
            self.assertIn("User-Agent", headers)
            self.assertEqual(timeout, 10.0)
            return FakeResponse()

        fake_module = types.SimpleNamespace(HfApi=FakeApi)
        fake_requests = types.SimpleNamespace(get=fake_get)
        with patch.dict("sys.modules", {"huggingface_hub": fake_module, "requests": fake_requests}):
            metadata = hf_models.fetch_gguf_quant_metadata("unsloth/GLM-5-GGUF")
        self.assertCountEqual(metadata.quantizations, ["Q4_1", "Q8_0"])
        self.assertAlmostEqual(metadata.vram_gb_by_quant["Q4_1"], 472.704, places=2)
        self.assertAlmostEqual(metadata.vram_gb_by_quant["Q8_0"], 801.345, places=2)

    def test_fetch_vllm_memory_breakdown_uses_hf_config(self) -> None:
        calls: list[tuple[str, str | None]] = []

        class FakeApi:
            def model_info(self, *, repo_id: str, revision: str | None, expand: list[str]):
                self._ = expand
                calls.append((repo_id, revision))
                return SimpleNamespace(
                    config={
                        "num_hidden_layers": 32,
                        "hidden_size": 4096,
                        "max_position_embeddings": 32768,
                    },
                    safetensors={"total": 16_000_000_000},
                    tags=["text-generation", "bf16"],
                )

        fake_module = types.SimpleNamespace(HfApi=FakeApi)
        with patch.dict("sys.modules", {"huggingface_hub": fake_module}):
            estimate = hf_models.fetch_vllm_memory_breakdown("Qwen/Qwen3-8B-Instruct")
            estimate_cached = hf_models.fetch_vllm_memory_breakdown("Qwen/Qwen3-8B-Instruct")

        assert estimate is not None
        assert estimate_cached is not None
        self.assertEqual(calls, [("Qwen/Qwen3-8B-Instruct", None)])
        self.assertGreater(estimate.total_gb, estimate.weights_gb)
        self.assertEqual(estimate.context_tokens, 32768)

    def test_fetch_vllm_memory_breakdown_uses_timeout_and_card_data_expand(self) -> None:
        calls: list[tuple[float, tuple[str, ...]]] = []

        class FakeApi:
            def model_info(
                self,
                *,
                repo_id: str,
                revision: str | None,
                timeout: float,
                expand: list[str] | None = None,
            ):
                self._ = (repo_id, revision)
                calls.append((timeout, tuple(expand or [])))
                return SimpleNamespace(
                    config={
                        "num_hidden_layers": 32,
                        "hidden_size": 4096,
                        "max_position_embeddings": 32768,
                    },
                    safetensors={"total": 16_000_000_000},
                    tags=["text-generation", "bf16"],
                    cardData={"context_length": 32768},
                )

        fake_module = types.SimpleNamespace(HfApi=FakeApi)
        with patch.dict("sys.modules", {"huggingface_hub": fake_module}):
            estimate = hf_models.fetch_vllm_memory_breakdown("Qwen/Qwen3-8B-Instruct")
        assert estimate is not None
        self.assertEqual(calls[0][0], hf_models._HF_REQUEST_TIMEOUT_SECONDS)
        self.assertIn("cardData", calls[0][1])

    def test_fetch_vllm_memory_breakdown_skips_repo_json_fetch_when_model_info_is_complete(self) -> None:
        class FakeApi:
            def model_info(self, *, repo_id: str, revision: str | None, expand: list[str]):
                self._ = (repo_id, revision, expand)
                return SimpleNamespace(
                    config={
                        "num_hidden_layers": 32,
                        "hidden_size": 4096,
                        "max_position_embeddings": 32768,
                    },
                    safetensors={"total": 16_000_000_000},
                    tags=["text-generation", "bf16"],
                    cardData={"context_length": 32768},
                )

        fake_module = types.SimpleNamespace(HfApi=FakeApi)
        with (
            patch.dict("sys.modules", {"huggingface_hub": fake_module}),
            patch(
                "llm_launchpad.core.hf_models._load_repo_json_file",
                side_effect=AssertionError("should not fetch repo JSON"),
            ),
        ):
            estimate = hf_models.fetch_vllm_memory_breakdown("Qwen/Qwen3-8B-Instruct")
        assert estimate is not None
        self.assertEqual(estimate.context_tokens, 32768)

    def test_fetch_vllm_memory_breakdown_uses_default_context_without_metadata(self) -> None:
        class FakeApi:
            def model_info(self, *, repo_id: str, revision: str | None, expand: list[str]):
                self._ = (repo_id, revision, expand)
                return SimpleNamespace(config={}, safetensors={}, tags=["text-generation", "fp8"])

        fake_module = types.SimpleNamespace(HfApi=FakeApi)
        with (
            patch.dict("sys.modules", {"huggingface_hub": fake_module}),
            patch("llm_launchpad.core.hf_models._load_repo_json_file", return_value=None),
        ):
            estimate = hf_models.fetch_vllm_memory_breakdown("Qwen/Qwen3-8B-FP8")
        assert estimate is not None
        self.assertEqual(estimate.context_tokens, hf_models._DEFAULT_CONTEXT_TOKENS)

    def test_fetch_vllm_memory_breakdown_retries_without_expand(self) -> None:
        class FakeApi:
            def __init__(self) -> None:
                self.calls = 0

            def model_info(self, *, repo_id: str, revision: str | None, expand=None):
                self.calls += 1
                if self.calls == 1:
                    raise RuntimeError("expand not supported")
                self._ = (repo_id, revision, expand)
                return SimpleNamespace(
                    config={"max_position_embeddings": 131072, "num_hidden_layers": 32, "hidden_size": 4096},
                    safetensors={"total": 16_000_000_000},
                    tags=["bf16"],
                )

        fake_module = types.SimpleNamespace(HfApi=FakeApi)
        with patch.dict("sys.modules", {"huggingface_hub": fake_module}):
            estimate = hf_models.fetch_vllm_memory_breakdown("Qwen/Qwen3-8B-Instruct")
        assert estimate is not None
        self.assertEqual(estimate.context_tokens, 131072)

    def test_extract_model_max_context_reads_nested_config(self) -> None:
        config = {
            "text_config": {"max_position_embeddings": 32768},
            "rope_scaling": {"original_max_position_embeddings": 65536},
            "tokenizer_config": {"model_max_length": 100000000000000000000},
        }
        value = hf_models._extract_model_max_context(config)
        self.assertEqual(value, 32768)

    def test_extract_model_max_context_handles_rope_scaling_factor(self) -> None:
        config = {
            "max_position_embeddings": 32768,
            "rope_scaling": {"factor": 8.0, "original_max_position_embeddings": 32768},
        }
        value = hf_models._extract_model_max_context(config)
        self.assertEqual(value, 262144)

    def test_extract_model_max_context_prefers_original_rope_base(self) -> None:
        config = {
            "max_position_embeddings": 131072,
            "rope_scaling": {"factor": 32.0, "original_max_position_embeddings": 4096},
        }
        value = hf_models._extract_model_max_context(config)
        self.assertEqual(value, 131072)

    def test_extract_model_max_context_parses_k_suffix(self) -> None:
        config = {"context_length": "200K"}
        value = hf_models._extract_model_max_context(config)
        self.assertEqual(value, 200000)

    def test_load_repo_json_file_falls_back_to_http_when_hf_hub_download_fails(self) -> None:
        def fake_hf_hub_download(*, repo_id: str, filename: str, revision: str | None, etag_timeout: float):
            self._ = (repo_id, filename, revision, etag_timeout)
            raise RuntimeError("network error")

        fake_module = types.SimpleNamespace(hf_hub_download=fake_hf_hub_download)
        with (
            patch.dict("sys.modules", {"huggingface_hub": fake_module}),
            patch(
                "llm_launchpad.core.hf_models._fetch_repo_json_file_via_http",
                return_value={"max_position_embeddings": 262144},
            ) as fetch_http,
        ):
            parsed = hf_models._load_repo_json_file("Qwen/Qwen3-Coder-Next", None, "config.json")
        assert parsed is not None
        self.assertEqual(parsed.get("max_position_embeddings"), 262144)
        self.assertEqual(fetch_http.call_count, 1)

    def test_load_repo_json_file_falls_back_to_http_when_downloaded_json_is_invalid(self) -> None:
        with NamedTemporaryFile("w", encoding="utf-8", delete=False) as handle:
            handle.write("{not-json")
            temp_path = Path(handle.name)

        def fake_hf_hub_download(*, repo_id: str, filename: str, revision: str | None, etag_timeout: float):
            self._ = (repo_id, filename, revision, etag_timeout)
            return str(temp_path)

        fake_module = types.SimpleNamespace(hf_hub_download=fake_hf_hub_download)
        with (
            patch.dict("sys.modules", {"huggingface_hub": fake_module}),
            patch(
                "llm_launchpad.core.hf_models._fetch_repo_json_file_via_http",
                return_value={"max_position_embeddings": 262144},
            ) as fetch_http,
        ):
            parsed = hf_models._load_repo_json_file("Qwen/Qwen3-Coder-Next", None, "config.json")
        try:
            temp_path.unlink(missing_ok=True)
        except Exception:
            pass
        assert parsed is not None
        self.assertEqual(parsed.get("max_position_embeddings"), 262144)
        self.assertEqual(fetch_http.call_count, 1)

    def test_fetch_vllm_memory_breakdown_uses_repo_json_context_when_model_info_missing(self) -> None:
        class FakeApi:
            def model_info(self, *, repo_id: str, revision: str | None, expand=None):
                self._ = (repo_id, revision, expand)
                return SimpleNamespace(config={}, safetensors={"total": 16_000_000_000}, tags=["bf16"])

        def fake_load(repo_id: str, revision: str | None, filename: str):
            self.assertEqual(repo_id, "zai-org/GLM-5")
            if filename == "config.json":
                return {
                    "num_hidden_layers": 64,
                    "hidden_size": 7168,
                    "max_position_embeddings": 32768,
                    "rope_scaling": {"factor": 8.0, "original_max_position_embeddings": 32768},
                }
            if filename == "tokenizer_config.json":
                return {"model_max_length": 262144}
            return None

        fake_module = types.SimpleNamespace(HfApi=FakeApi)
        with (
            patch.dict("sys.modules", {"huggingface_hub": fake_module}),
            patch("llm_launchpad.core.hf_models._load_repo_json_file", side_effect=fake_load),
        ):
            estimate = hf_models.fetch_vllm_memory_breakdown("zai-org/GLM-5")

        assert estimate is not None
        self.assertEqual(estimate.context_tokens, 262144)

    def test_fetch_vllm_memory_breakdown_uses_card_data_context(self) -> None:
        class FakeApi:
            def model_info(self, *, repo_id: str, revision: str | None, expand=None):
                self._ = (repo_id, revision, expand)
                return SimpleNamespace(
                    config={},
                    cardData={"context_length": "256K"},
                    safetensors={"total": 16_000_000_000},
                    tags=["bf16"],
                )

        fake_module = types.SimpleNamespace(HfApi=FakeApi)
        with (
            patch.dict("sys.modules", {"huggingface_hub": fake_module}),
            patch("llm_launchpad.core.hf_models._load_repo_json_file", return_value=None),
        ):
            estimate = hf_models.fetch_vllm_memory_breakdown("zai-org/GLM-5")
        assert estimate is not None
        self.assertEqual(estimate.context_tokens, 256000)


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

import unittest
from types import SimpleNamespace
from unittest.mock import patch

from llm_launchpad.core.backend import ModalBackend, _extract_modal_app_rows
from llm_launchpad.core.orchestrator import Orchestrator
from llm_launchpad.protocol.enums import BackendType
from llm_launchpad.protocol.events import LogEvent, OperationCompleteEvent
from llm_launchpad.protocol.models import DeploymentConfig, EndpointInfo, LaunchpadSettings


class BackendMultiInstanceTests(unittest.TestCase):
    def test_extract_modal_app_rows_infers_backend_for_multi_instance_names(self) -> None:
        payload = [
            {"name": "vllm-qwen3", "app_id": "ap-1", "state": "running"},
            {"name": "llamacpp-qwen2-5", "app_id": "ap-2", "state": "running"},
        ]
        rows = _extract_modal_app_rows(payload)
        self.assertEqual(len(rows), 2)
        self.assertEqual(rows[0].backend, BackendType.VLLM)
        self.assertEqual(rows[0].instance_name, "qwen3")
        self.assertEqual(rows[1].backend, BackendType.LLAMACPP)
        self.assertEqual(rows[1].instance_name, "qwen2-5")

    def test_default_server_url_accepts_explicit_app_name(self) -> None:
        url = ModalBackend.default_server_url("alice", app_name="vllm-qwen3")
        self.assertEqual(url, "https://alice--vllm-qwen3-serve.modal.run")

    def test_default_server_url_accepts_function_slug(self) -> None:
        url = ModalBackend.default_server_url(
            "alice",
            app_name="vllm-qwen3",
            function_slug="alpha-bravo",
        )
        self.assertEqual(url, "https://alice--vllm-qwen3-serve-alpha-bravo.modal.run")

    def test_extract_modal_web_url_parses_deploy_line(self) -> None:
        line = (
            "└── 🔨 Created web function serve-discerning-tapir => "
            "https://alice--vllm-minimax-m2-5-serve-disc-42c728.modal.run (label truncated)"
        )
        parsed = ModalBackend.extract_modal_web_url(line)
        self.assertEqual(parsed, "https://alice--vllm-minimax-m2-5-serve-disc-42c728.modal.run")


class OrchestratorMultiInstanceTests(unittest.TestCase):
    def test_deploy_assigns_and_forwards_function_slug_for_non_vllm(self) -> None:
        orch = Orchestrator(config_store=SimpleNamespace(load=lambda: LaunchpadSettings()))
        captured_env: list[dict[str, str] | None] = []

        def _fake_run_streaming(command: list[str], env=None):  # type: ignore[no-untyped-def]
            captured_env.append(env)
            yield OperationCompleteEvent(success=True, exit_code=0)

        config = DeploymentConfig(
            backend=BackendType.LLAMACPP,
            app_name="llamacpp-qwen3",
            do_deploy=True,
        )
        with patch("llm_launchpad.core.orchestrator.random_function_slug", return_value="alpha-bravo"):
            with patch("llm_launchpad.core.backend.ModalBackend.run_streaming", side_effect=_fake_run_streaming):
                list(orch.deploy(config))
        self.assertEqual(config.function_slug, "alpha-bravo")
        self.assertTrue(captured_env)
        self.assertEqual(captured_env[0]["MODAL_FUNCTION_SLUG"], "alpha-bravo")

    def test_deploy_does_not_auto_assign_function_slug_for_vllm(self) -> None:
        orch = Orchestrator(config_store=SimpleNamespace(load=lambda: LaunchpadSettings()))
        captured_env: list[dict[str, str] | None] = []

        def _fake_run_streaming(command: list[str], env=None):  # type: ignore[no-untyped-def]
            captured_env.append(env)
            yield OperationCompleteEvent(success=True, exit_code=0)

        config = DeploymentConfig(
            backend=BackendType.VLLM,
            app_name="vllm-qwen3",
            do_deploy=True,
        )
        with patch("llm_launchpad.core.orchestrator.random_function_slug", return_value="alpha-bravo"):
            with patch("llm_launchpad.core.backend.ModalBackend.run_streaming", side_effect=_fake_run_streaming):
                list(orch.deploy(config))
        self.assertIsNone(config.function_slug)
        self.assertTrue(captured_env)
        self.assertNotIn("MODAL_FUNCTION_SLUG", captured_env[0] or {})

    def test_tail_logs_targets_explicit_app_name(self) -> None:
        orch = Orchestrator()
        captured: list[list[str]] = []

        def _fake_run_streaming(command: list[str], env=None):  # type: ignore[no-untyped-def]
            captured.append(command)
            if False:
                yield None
            return
            yield  # pragma: no cover

        with patch("llm_launchpad.core.backend.ModalBackend.logs_follow_args", return_value=[]):
            with patch("llm_launchpad.core.backend.ModalBackend.run_streaming", side_effect=_fake_run_streaming):
                list(orch.tail_logs(BackendType.VLLM, follow=False, app_name="vllm-qwen3"))
        self.assertTrue(captured)
        self.assertEqual(captured[0], ["modal", "app", "logs", "vllm-qwen3"])

    def test_tail_logs_prefers_explicit_app_id(self) -> None:
        orch = Orchestrator()
        captured: list[list[str]] = []

        def _fake_run_streaming(command: list[str], env=None):  # type: ignore[no-untyped-def]
            captured.append(command)
            if False:
                yield None
            return
            yield  # pragma: no cover

        with patch("llm_launchpad.core.backend.ModalBackend.logs_follow_args", return_value=[]):
            with patch("llm_launchpad.core.backend.ModalBackend.run_streaming", side_effect=_fake_run_streaming):
                list(orch.tail_logs(BackendType.VLLM, follow=False, app_name="vllm-qwen3", app_id="ap-123"))
        self.assertTrue(captured)
        self.assertEqual(captured[0], ["modal", "app", "logs", "ap-123"])

    def test_stop_app_targets_explicit_app_name(self) -> None:
        orch = Orchestrator()
        captured: list[list[str]] = []

        def _fake_run_streaming(command: list[str], env=None):  # type: ignore[no-untyped-def]
            captured.append(command)
            if False:
                yield None
            return
            yield  # pragma: no cover

        with patch("llm_launchpad.core.backend.ModalBackend.run_streaming", side_effect=_fake_run_streaming):
            events = list(orch.stop_app(BackendType.VLLM, app_name="vllm-qwen2-5"))
        self.assertTrue(captured)
        self.assertEqual(captured[0], ["modal", "app", "stop", "vllm-qwen2-5"])
        self.assertTrue(any(isinstance(event, LogEvent) for event in events))

    def test_stop_app_prefers_explicit_app_id(self) -> None:
        orch = Orchestrator()
        captured: list[list[str]] = []

        def _fake_run_streaming(command: list[str], env=None):  # type: ignore[no-untyped-def]
            captured.append(command)
            if False:
                yield None
            return
            yield  # pragma: no cover

        with patch("llm_launchpad.core.backend.ModalBackend.run_streaming", side_effect=_fake_run_streaming):
            list(orch.stop_app(BackendType.VLLM, app_name="vllm-qwen2-5", app_id="ap-123"))
        self.assertTrue(captured)
        self.assertEqual(captured[0], ["modal", "app", "stop", "ap-123"])

    def test_list_deployments_handles_empty_json_as_success(self) -> None:
        orch = Orchestrator()
        with patch("llm_launchpad.core.backend.ModalBackend.list_apps", return_value=[]):
            with patch("llm_launchpad.core.backend.ModalBackend.list_apps_raw") as raw_mock:
                events = list(orch.list_deployments())
        raw_mock.assert_not_called()
        self.assertTrue(
            any(isinstance(event, LogEvent) and event.line == "No launchpad deployments found." for event in events)
        )
        self.assertTrue(
            any(
                isinstance(event, OperationCompleteEvent)
                and event.success
                and isinstance(event.data, list)
                and len(event.data) == 0
                for event in events
            )
        )

    def test_list_deployments_hides_historical_duplicates_for_same_app(self) -> None:
        orch = Orchestrator()
        rows = [
            EndpointInfo(
                name="llamacpp-edge-quant-nanbeige4",
                app_id="ap-live",
                state="deployed",
                backend=BackendType.LLAMACPP,
                instance_name="edge-quant-nanbeige4",
            ),
            EndpointInfo(
                name="llamacpp-edge-quant-nanbeige4",
                app_id="ap-old-1",
                state="stopped",
                backend=BackendType.LLAMACPP,
                instance_name="edge-quant-nanbeige4",
            ),
            EndpointInfo(
                name="llamacpp-edge-quant-nanbeige4",
                app_id="ap-old-2",
                state="stopped",
                backend=BackendType.LLAMACPP,
                instance_name="edge-quant-nanbeige4",
            ),
            EndpointInfo(
                name="llamacpp-server",
                app_id="ap-default",
                state="stopped",
                backend=BackendType.LLAMACPP,
                instance_name="default",
            ),
        ]

        with patch("llm_launchpad.core.backend.ModalBackend.list_apps", return_value=rows):
            events = list(orch.list_deployments())

        log_lines = [event.line for event in events if isinstance(event, LogEvent)]
        deployment_lines = [line for line in log_lines if line.startswith("  backend=")]
        self.assertEqual(len(deployment_lines), 2)
        self.assertTrue(any("ap-live" in line and "state=deployed" in line for line in deployment_lines))
        self.assertFalse(any("ap-old-1" in line for line in deployment_lines))
        self.assertFalse(any("ap-old-2" in line for line in deployment_lines))
        self.assertTrue(any("hidden 2 historical duplicate app rows" in line for line in log_lines))

        done = next(
            event
            for event in events
            if isinstance(event, OperationCompleteEvent) and event.operation.value == "list"
        )
        self.assertTrue(done.success)
        self.assertIsInstance(done.data, list)
        self.assertEqual(len(done.data), 2)

    def test_list_deployments_keeps_concurrent_active_duplicates(self) -> None:
        orch = Orchestrator()
        rows = [
            EndpointInfo(
                name="llamacpp-glm5-rtxpro",
                app_id="ap-first",
                state="ephemeral",
                backend=BackendType.LLAMACPP,
                instance_name="glm5-rtxpro",
            ),
            EndpointInfo(
                name="llamacpp-glm5-rtxpro",
                app_id="ap-second",
                state="ephemeral",
                backend=BackendType.LLAMACPP,
                instance_name="glm5-rtxpro",
            ),
            EndpointInfo(
                name="llamacpp-glm5-rtxpro",
                app_id="ap-old",
                state="stopped",
                backend=BackendType.LLAMACPP,
                instance_name="glm5-rtxpro",
            ),
        ]

        with patch("llm_launchpad.core.backend.ModalBackend.list_apps", return_value=rows):
            events = list(orch.list_deployments())

        log_lines = [event.line for event in events if isinstance(event, LogEvent)]
        deployment_lines = [line for line in log_lines if line.startswith("  backend=")]
        self.assertEqual(len(deployment_lines), 2)
        self.assertTrue(any("ap-first" in line and "state=ephemeral" in line for line in deployment_lines))
        self.assertTrue(any("ap-second" in line and "state=ephemeral" in line for line in deployment_lines))
        self.assertFalse(any("ap-old" in line for line in deployment_lines))


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

import types
import unittest
from unittest.mock import patch

from llm_launchpad.core import modal_gpu


_DOC_SNIPPET = """
## Specifying GPU type

Modal supports the following values for this parameter:

* `T4`
* `L4`
* `A10`
* `A100`
* `A100-40GB`
* `A100-80GB`
* `L40S`
* `RTX-PRO-6000`
* `H100`/`H100!`
* `H200`
* `B200`
* `B200+`

For instance, to run a Function with eight H100s:

@app.function(gpu="H100:8")
def run_llama_405b_fp8():
    ...

## Specifying GPU count
"""

_PRICING_SNIPPET = """
### Compute costs

GPU Tasks

Nvidia B200

$0.001736 / sec

Nvidia H200

$0.001261 / sec

Nvidia H100

$0.001097 / sec

Nvidia RTX PRO 6000

$0.000842 / sec

Nvidia A100, 80 GB

$0.000694 / sec

Nvidia A100, 40 GB

$0.000583 / sec

Nvidia L40S

$0.000542 / sec

Nvidia A10

$0.000306 / sec

Nvidia L4

$0.000222 / sec

Nvidia T4

$0.000164 / sec

CPU
"""


class ModalGpuTypesTests(unittest.TestCase):
    def test_parse_modal_gpu_types_from_docs_snippet(self) -> None:
        parsed = modal_gpu._parse_modal_gpu_types(_DOC_SNIPPET)
        self.assertEqual(
            parsed,
            [
                "T4",
                "L4",
                "A10",
                "A100",
                "A100-40GB",
                "A100-80GB",
                "L40S",
                "RTX-PRO-6000",
                "H100",
                "H100!",
                "H200",
                "B200",
                "B200+",
            ],
        )

    def test_fetch_modal_gpu_types_reads_docs_response(self) -> None:
        class FakeResponse:
            status_code = 200
            text = _DOC_SNIPPET

        def fake_get(url: str, timeout: float, headers: dict[str, str]):
            self.assertEqual(url, modal_gpu.MODAL_GPU_GUIDE_URL)
            self.assertEqual(timeout, 7.0)
            self.assertIn("User-Agent", headers)
            return FakeResponse()

        fake_requests = types.SimpleNamespace(get=fake_get)
        with patch.dict("sys.modules", {"requests": fake_requests}):
            values = modal_gpu.fetch_modal_gpu_types(timeout=7.0)
        self.assertIn("H100", values)
        self.assertIn("A100-80GB", values)
        self.assertIn("RTX-PRO-6000", values)
        self.assertIn("B200+", values)
        self.assertNotIn("H100:8", values)

    def test_parse_modal_gpu_pricing_from_pricing_snippet(self) -> None:
        pricing = modal_gpu._parse_modal_gpu_pricing(_PRICING_SNIPPET)
        self.assertAlmostEqual(pricing["B200"], 6.2496)
        self.assertAlmostEqual(pricing["H200"], 4.5396)
        self.assertAlmostEqual(pricing["H100"], 3.9492)
        self.assertAlmostEqual(pricing["RTX-PRO-6000"], 3.0312)
        self.assertAlmostEqual(pricing["A100-80GB"], 2.4984)
        self.assertAlmostEqual(pricing["A100-40GB"], 2.0988)
        self.assertAlmostEqual(pricing["L40S"], 1.9512)
        self.assertAlmostEqual(pricing["A10"], 1.1016)
        self.assertAlmostEqual(pricing["L4"], 0.7992)
        self.assertAlmostEqual(pricing["T4"], 0.5904)
        self.assertAlmostEqual(pricing["A100"], pricing["A100-40GB"])
        self.assertAlmostEqual(pricing["H100!"], pricing["H100"])
        self.assertAlmostEqual(pricing["B200+"], pricing["B200"])

    def test_fetch_modal_gpu_catalog_merges_docs_and_pricing(self) -> None:
        class FakeResponse:
            def __init__(self, text: str) -> None:
                self.status_code = 200
                self.text = text

        def fake_get(url: str, timeout: float, headers: dict[str, str]):
            self.assertEqual(timeout, 7.0)
            self.assertIn("User-Agent", headers)
            if url == modal_gpu.MODAL_GPU_GUIDE_URL:
                return FakeResponse(_DOC_SNIPPET)
            if url == modal_gpu.MODAL_PRICING_URL:
                return FakeResponse(_PRICING_SNIPPET)
            self.fail(f"Unexpected URL: {url}")

        fake_requests = types.SimpleNamespace(get=fake_get)
        with patch.dict("sys.modules", {"requests": fake_requests}):
            catalog = modal_gpu.fetch_modal_gpu_catalog(timeout=7.0)

        pricing_by_gpu = {entry.value: entry.price_per_hour_usd for entry in catalog}
        self.assertAlmostEqual(pricing_by_gpu["B200"] or 0.0, 6.2496)
        self.assertAlmostEqual(pricing_by_gpu["B200+"] or 0.0, 6.2496)
        self.assertAlmostEqual(pricing_by_gpu["H100"] or 0.0, 3.9492)
        self.assertAlmostEqual(pricing_by_gpu["H100!"] or 0.0, 3.9492)
        self.assertAlmostEqual(pricing_by_gpu["A100"] or 0.0, 2.0988)
        self.assertAlmostEqual(pricing_by_gpu["A100-80GB"] or 0.0, 2.4984)
        self.assertAlmostEqual(pricing_by_gpu["RTX-PRO-6000"] or 0.0, 3.0312)

    def test_fetch_modal_gpu_catalog_keeps_types_when_pricing_fetch_fails(self) -> None:
        class FakeResponse:
            def __init__(self, status_code: int, text: str) -> None:
                self.status_code = status_code
                self.text = text

        def fake_get(url: str, timeout: float, headers: dict[str, str]):
            self.assertEqual(timeout, 7.0)
            self.assertIn("User-Agent", headers)
            if url == modal_gpu.MODAL_GPU_GUIDE_URL:
                return FakeResponse(200, _DOC_SNIPPET)
            if url == modal_gpu.MODAL_PRICING_URL:
                return FakeResponse(503, "")
            self.fail(f"Unexpected URL: {url}")

        fake_requests = types.SimpleNamespace(get=fake_get)
        with patch.dict("sys.modules", {"requests": fake_requests}):
            catalog = modal_gpu.fetch_modal_gpu_catalog(timeout=7.0)

        self.assertEqual(catalog[0].value, "T4")
        self.assertIsNone(catalog[0].price_per_hour_usd)

    def test_fetch_modal_gpu_types_raises_on_http_error(self) -> None:
        class FakeResponse:
            status_code = 404
            text = ""

        fake_requests = types.SimpleNamespace(get=lambda *_args, **_kwargs: FakeResponse())
        with patch.dict("sys.modules", {"requests": fake_requests}):
            with self.assertRaisesRegex(RuntimeError, "HTTP 404"):
                modal_gpu.fetch_modal_gpu_types()


if __name__ == "__main__":
    unittest.main()



"""Manage endpoints screen: list, status, logs, stop."""

from __future__ import annotations

from textual import events
from textual.app import ComposeResult
from textual.binding import Binding
from textual.containers import Vertical
from textual.widgets import Footer, Input, OptionList, Static, Switch
from textual.widgets.option_list import Option

from ...protocol.models import EndpointInfo
from ..widgets.input_form import FormField, ToggleField
from .copy_enabled import CopyEnabledScreen


def _is_stoppable_state(state: str) -> bool:
    """Return True when the app state should appear in the manage UI."""
    return state.strip().lower() in {
        "building",
        "deployed",
        "deploying",
        "ephemeral",
        "initializing",
        "pending",
        "queued",
        "running",
        "starting",
    }


def _build_backend_app_options(
    instances: list[EndpointInfo],
) -> tuple[list[Option], dict[str, EndpointInfo]]:
    """Build options and map option IDs to the originating Modal row."""
    options: list[Option] = []
    option_to_target: dict[str, EndpointInfo] = {}

    for index, row in enumerate(instances):
        if row.backend is None:
            continue
        option_id = f"app-id:{row.app_id}" if row.app_id else f"app-name:{row.name}:{index}"
        label = f"  [{row.backend.value}] {row.name}  ({row.state}"
        if row.app_id:
            label += f", {row.app_id}"
        label += ")"
        options.append(Option(label, id=option_id))
        option_to_target[option_id] = row

    return options, option_to_target


class ManageScreen(CopyEnabledScreen):
    """Manage endpoints: pick action, then pick backend and params."""

    BINDINGS = [
        Binding("escape", "pop_screen", "Back", show=True),
    ]

    def compose(self) -> ComposeResult:
        with Vertical(id="menu-container"):
            yield Static("[bold #7bf168]Manage Endpoints[/]")
            yield Static("")
            yield OptionList(
                Option("  List apps                 Show active launchpad Modal apps", id="list"),
                Option("  Status check              Probe endpoint health", id="status"),
                Option("  Tail logs                 Stream Modal app logs", id="logs"),
                Option("  Stop app                  Stop an active Modal app", id="stop"),
                id="manage-action-list",
            )
        yield Footer()

    def on_mount(self) -> None:
        action_list = self.query_one("#manage-action-list", OptionList)
        if action_list.option_count > 0:
            action_list.highlighted = 0
        action_list.focus()

    def on_option_list_option_selected(self, event: OptionList.OptionSelected) -> None:
        opt = event.option.id
        if opt == "list":
            self.app.begin_list()  # type: ignore[attr-defined]
        elif opt == "status":
            self.app.push_screen(StatusParamsScreen())
        elif opt == "logs":
            self.app.push_screen(LogsParamsScreen())
        elif opt == "stop":
            self.app.push_screen(StopParamsScreen())

    def action_pop_screen(self) -> None:
        self.app.pop_screen()


class StatusParamsScreen(CopyEnabledScreen):
    """Params for status check: backend, optional URL override, timeout."""

    BINDINGS = [
        Binding("escape", "pop_screen", "Back", show=True),
        Binding("enter", "do_submit", "Submit", show=True),
    ]

    def compose(self) -> ComposeResult:
        with Vertical(id="menu-container"):
            yield Static("[bold #7bf168]Status Check[/]")
            yield Static("")
            yield Static("[bold]Choose active app[/bold]")
            yield OptionList(id="status-instance-list")
            yield FormField(
                "Server URL override (optional)",
                "status-url",
                hint="Leave blank for default",
            )
            yield FormField("Timeout (seconds)", "status-timeout", default="60")
            yield Static("")
        yield Footer()

    def on_mount(self) -> None:
        self._selected_row: EndpointInfo | None = None
        self._row_by_option_id: dict[str, EndpointInfo] = {}
        self._load_instances()
        self.query_one("#status-instance-list", OptionList).focus()

    def on_option_list_option_selected(self, event: OptionList.OptionSelected) -> None:
        if event.option_list.id == "status-instance-list":
            selected = str(event.option.id)
            row = self._row_by_option_id.get(selected)
            if row is None:
                return
            self._selected_row = row
            self._submit()

    def action_do_submit(self) -> None:
        self._submit()

    def _submit(self) -> None:
        if self._selected_row is None or self._selected_row.backend is None:
            return
        url = self.query_one("#status-url", Input).value.strip() or self._selected_row.web_url or None
        timeout_str = self.query_one("#status-timeout", Input).value.strip()
        try:
            timeout = int(timeout_str) if timeout_str else 60
        except ValueError:
            timeout = 60
        self.app.begin_status(  # type: ignore[attr-defined]
            self._selected_row.backend,
            url,
            timeout,
            app_name=self._selected_row.name,
            served_model_name=self._selected_row.served_model_name,
        )

    def _load_instances(self) -> None:
        instance_list = self.query_one("#status-instance-list", OptionList)
        instances = self.app.list_instances()  # type: ignore[attr-defined]
        checkable_instances = [row for row in instances if _is_stoppable_state(row.state)]
        if not checkable_instances:
            self._row_by_option_id = {}
            instance_list.set_options([Option("  No active apps found")])
            if instance_list.option_count > 0:
                instance_list.highlighted = 0
            return
        options, self._row_by_option_id = _build_backend_app_options(checkable_instances)
        instance_list.set_options(options)
        if options:
            instance_list.highlighted = 0

    def action_pop_screen(self) -> None:
        self.app.pop_screen()


class LogsParamsScreen(CopyEnabledScreen):
    """Params for log tailing: backend, follow toggle."""

    BINDINGS = [
        Binding("escape", "pop_screen", "Back", show=True),
    ]

    def compose(self) -> ComposeResult:
        with Vertical(id="menu-container"):
            yield Static("[bold #7bf168]Tail Logs[/]")
            yield Static("")
            yield Static("[bold]Choose active app[/bold]")
            yield OptionList(id="logs-instance-list")
            yield ToggleField("Follow log stream", "logs-follow", default=True)
        yield Footer()

    def on_mount(self) -> None:
        self._selected_row: EndpointInfo | None = None
        self._target_by_option_id: dict[str, EndpointInfo] = {}
        self._load_instances()
        self.query_one("#logs-instance-list", OptionList).focus()

    def on_option_list_option_selected(self, event: OptionList.OptionSelected) -> None:
        if event.option_list.id == "logs-instance-list":
            selected = str(event.option.id)
            self._selected_row = self._target_by_option_id.get(selected)
            if self._selected_row is None:
                return
            self._submit()

    def _submit(self) -> None:
        if self._selected_row is None or self._selected_row.backend is None:
            return
        follow = self.query_one("#logs-follow", Switch).value
        self.app.begin_logs(  # type: ignore[attr-defined]
            self._selected_row.backend,
            follow,
            app_name=self._selected_row.name,
            app_id=self._selected_row.app_id or None,
        )

    def _load_instances(self) -> None:
        instance_list = self.query_one("#logs-instance-list", OptionList)
        instances = self.app.list_instances()  # type: ignore[attr-defined]
        loggable_instances = [row for row in instances if _is_stoppable_state(row.state)]
        if not loggable_instances:
            self._target_by_option_id = {}
            instance_list.set_options([Option("  No active apps found")])
            if instance_list.option_count > 0:
                instance_list.highlighted = 0
            return
        options, self._target_by_option_id = _build_backend_app_options(loggable_instances)
        instance_list.set_options(options)
        if options:
            instance_list.highlighted = 0

    def action_pop_screen(self) -> None:
        self.app.pop_screen()


class StopParamsScreen(CopyEnabledScreen):
    """Confirmation for stopping a backend."""

    BINDINGS = [
        Binding("escape", "pop_screen", "Back", show=True),
        Binding("enter", "do_submit", "Stop", show=True),
    ]

    def compose(self) -> ComposeResult:
        with Vertical(id="menu-container"):
            yield Static("[bold #7bf168]Stop App[/]")
            yield Static("")
            yield Static("[bold]Choose active app[/bold]")
            yield OptionList(id="stop-instance-list")
            yield Static("")
            yield Static("[yellow]Warning:[/yellow] This will stop the selected Modal app.")
            yield Static("[dim]Select an app to confirm and stop.[/dim]")
        yield Footer()

    def on_option_list_option_selected(self, event: OptionList.OptionSelected) -> None:
        if event.option_list.id == "stop-instance-list":
            selected = str(event.option.id)
            row = self._target_by_option_id.get(selected)
            if row is None or row.backend is None:
                return
            self.app.begin_stop(  # type: ignore[attr-defined]
                row.backend,
                app_name=row.name,
                app_id=row.app_id or None,
            )

    def on_mount(self) -> None:
        self._target_by_option_id: dict[str, EndpointInfo] = {}
        self._was_suspended = False
        self._load_instances()
        self.query_one("#stop-instance-list", OptionList).focus()

    def on_screen_suspend(self, _: events.ScreenSuspend) -> None:
        self._was_suspended = True

    def on_screen_resume(self, _: events.ScreenResume) -> None:
        """Refresh instance list when returning from the stop monitor screen."""
        if self._was_suspended:
            self._was_suspended = False
            self._load_instances()

    def action_do_submit(self) -> None:
        instance_list = self.query_one("#stop-instance-list", OptionList)
        highlighted = instance_list.highlighted_option
        if highlighted is None:
            return
        selected = str(highlighted.id)
        row = self._target_by_option_id.get(selected)
        if row is None or row.backend is None:
            return
        self.app.begin_stop(  # type: ignore[attr-defined]
            row.backend,
            app_name=row.name,
            app_id=row.app_id or None,
        )

    def _load_instances(self) -> None:
        instance_list = self.query_one("#stop-instance-list", OptionList)
        instances = self.app.list_instances()  # type: ignore[attr-defined]
        stoppable_instances = [row for row in instances if _is_stoppable_state(row.state)]
        if not stoppable_instances:
            self._target_by_option_id = {}
            instance_list.set_options([Option("  No active apps found")])
            if instance_list.option_count > 0:
                instance_list.highlighted = 0
            return
        options, self._target_by_option_id = _build_backend_app_options(stoppable_instances)
        instance_list.set_options(options)
        if options:
            instance_list.highlighted = 0

    def action_pop_screen(self) -> None:
        self.app.pop_screen()



from __future__ import annotations

import unittest
from unittest.mock import patch

from llm_launchpad.core.hf_auth import (
    _extract_username,
    get_huggingface_auth_status,
)


class _LocalTokenNotFoundError(Exception):
    pass


class _InvalidTokenError(Exception):
    pass


class HuggingFaceAuthTests(unittest.TestCase):
    def test_extract_username_reads_common_whoami_shapes(self) -> None:
        self.assertEqual(_extract_username({"name": "alice"}), "alice")
        self.assertEqual(_extract_username({"user": {"name": "bob"}}), "bob")
        self.assertEqual(_extract_username({"auth": {"username": "carol"}}), "carol")

    @patch("llm_launchpad.core.hf_auth._has_local_token", return_value=False)
    def test_get_huggingface_auth_status_returns_unauthenticated_without_token(self, _mock_token) -> None:
        status = get_huggingface_auth_status()
        self.assertFalse(status.authenticated)
        self.assertIsNone(status.username)
        self.assertIsNone(status.error)

    @patch("llm_launchpad.core.hf_auth._has_local_token", return_value=True)
    @patch("huggingface_hub.HfApi")
    def test_get_huggingface_auth_status_returns_username_on_success(self, mock_hf_api, _mock_token) -> None:  # type: ignore[no-untyped-def]
        mock_hf_api.return_value.whoami.return_value = {"name": "alice"}

        status = get_huggingface_auth_status()

        self.assertTrue(status.authenticated)
        self.assertEqual(status.username, "alice")
        self.assertIsNone(status.error)

    @patch("llm_launchpad.core.hf_auth._has_local_token", return_value=True)
    @patch("huggingface_hub.HfApi")
    def test_get_huggingface_auth_status_handles_missing_login(self, mock_hf_api, _mock_token) -> None:  # type: ignore[no-untyped-def]
        mock_hf_api.return_value.whoami.side_effect = _LocalTokenNotFoundError("Run `hf auth login`")

        status = get_huggingface_auth_status()

        self.assertFalse(status.authenticated)
        self.assertIsNone(status.username)
        self.assertIsNone(status.error)

    @patch("llm_launchpad.core.hf_auth._has_local_token", return_value=True)
    @patch("huggingface_hub.HfApi")
    def test_get_huggingface_auth_status_handles_invalid_token(self, mock_hf_api, _mock_token) -> None:  # type: ignore[no-untyped-def]
        mock_hf_api.return_value.whoami.side_effect = _InvalidTokenError("401 Unauthorized: Invalid user token")

        status = get_huggingface_auth_status()

        self.assertFalse(status.authenticated)
        self.assertEqual(status.error, "Invalid Hugging Face token")


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

import unittest

from llm_launchpad.core.naming import (
    auto_instance_name_for_backend,
    build_app_name,
    default_served_model_name,
    infer_backend_from_app_name,
    infer_instance_from_app_name,
    legacy_app_name,
    modal_function_name,
    slugify_instance_name,
)
from llm_launchpad.protocol.enums import BackendType


class NamingTests(unittest.TestCase):
    def test_slugify_instance_name_normalizes_model_ids(self) -> None:
        slug = slugify_instance_name("Qwen/Qwen2.5-Coder_7B-Instruct")
        self.assertEqual(slug, "qwen-qwen2-5-coder-7b-instruct")

    def test_auto_instance_name_uses_model_hint(self) -> None:
        name = auto_instance_name_for_backend(BackendType.VLLM, "meta-llama/Llama-3.1-8B-Instruct")
        self.assertEqual(name, "meta-llama-llama-3-1-8b-instruct")

    def test_default_served_model_name_uses_model_id_suffix(self) -> None:
        alias = default_served_model_name("Qwen/Qwen3-0.6B")
        self.assertEqual(alias, "Qwen3-0.6B")

    def test_default_served_model_name_falls_back_to_default(self) -> None:
        self.assertEqual(default_served_model_name(""), "llm")
        self.assertEqual(default_served_model_name(None), "llm")

    def test_build_app_name(self) -> None:
        self.assertEqual(build_app_name(BackendType.VLLM, "qwen3"), "vllm-qwen3")
        self.assertEqual(build_app_name(BackendType.LLAMACPP, "qwen25"), "llamacpp-qwen25")

    def test_infer_backend_and_instance_from_app_name(self) -> None:
        app_name = "vllm-qwen3-coder"
        backend = infer_backend_from_app_name(app_name)
        self.assertEqual(backend, BackendType.VLLM)
        self.assertEqual(infer_instance_from_app_name(app_name, backend), "qwen3-coder")

    def test_legacy_app_name_roundtrip(self) -> None:
        app_name = legacy_app_name(BackendType.LLAMACPP)
        backend = infer_backend_from_app_name(app_name)
        self.assertEqual(backend, BackendType.LLAMACPP)
        self.assertEqual(infer_instance_from_app_name(app_name, backend), "default")

    def test_modal_function_name_appends_slug(self) -> None:
        self.assertEqual(modal_function_name("serve", "alpha-bravo"), "serve-alpha-bravo")
        self.assertEqual(modal_function_name("serve", None), "serve")


if __name__ == "__main__":
    unittest.main()



"""Data models shared across the launchpad stack."""

from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Dict, Optional

from .enums import BackendType


@dataclass
class LaunchpadSettings:
    """Persisted user settings (scaledown, etc.)."""

    scaledown_window: int = 1800  # seconds (30 minutes)

    def to_env(self) -> Dict[str, str]:
        """Derive Modal environment variables from settings."""
        env: Dict[str, str] = {}
        if self.scaledown_window > 0:
            env["SCALEDOWN_WINDOW"] = str(self.scaledown_window)
        return env

    def to_dict(self) -> Dict[str, Any]:
        return {"SCALEDOWN_WINDOW": self.scaledown_window}

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> LaunchpadSettings:
        raw_value = data.get("SCALEDOWN_WINDOW", data.get("scaledown_window", 1800))
        return cls(scaledown_window=int(raw_value))


@dataclass
class DeploymentConfig:
    """All parameters needed to execute a deployment."""

    backend: BackendType = BackendType.LLAMACPP

    # llama.cpp specific
    preset: Optional[str] = None
    repo_id: Optional[str] = None
    quant: Optional[str] = None
    revision: Optional[str] = None
    preload: bool = True
    server_args: Optional[str] = None
    host: Optional[str] = None
    port: Optional[int] = None
    n_gpu_layers: Optional[int] = None
    llamacpp_image_no_cache: Optional[bool] = None
    gpu_type: Optional[str] = None
    gpu_count: Optional[int] = None

    # vLLM specific
    model_name: Optional[str] = None
    model_revision: Optional[str] = None
    served_model_name: Optional[str] = None
    fast_boot: Optional[bool] = None
    n_gpu: Optional[int] = None
    trust_remote_code: Optional[bool] = None
    reasoning_parser: Optional[str] = None
    tool_call_parser: Optional[str] = None
    default_chat_template_kwargs: Optional[str] = None

    # Deployment options
    do_deploy: bool = True
    run_smoke: bool = False
    do_warmup: bool = True
    show_debug_logs: bool = False

    # Instance identity
    instance_name: Optional[str] = None
    app_name: Optional[str] = None
    function_slug: Optional[str] = None


@dataclass
class EndpointInfo:
    """A single row from `modal app list`."""

    name: str = ""
    app_id: str = ""
    state: str = "unknown"
    backend: Optional[BackendType] = None
    instance_name: Optional[str] = None
    web_url: Optional[str] = None
    served_model_name: Optional[str] = None
    display_name: Optional[str] = None
    model_name: Optional[str] = None
    repo_id: Optional[str] = None
    quant: Optional[str] = None
    runtime_status: Optional[str] = None
    runtime_status_detail: Optional[str] = None


@dataclass
class StoredModelInfo:
    """Single cached model entry in Modal storage."""

    backend: BackendType
    model_id: str
    revision: Optional[str] = None
    quant: Optional[str] = None
    size_bytes: int = 0
    file_count: int = 0
    source_volume: str = ""
    paths: list[str] | None = None
    incomplete: bool = False


@dataclass
class StorageSnapshot:
    """Cached model inventory grouped by backend."""

    llamacpp_models: list[StoredModelInfo]
    vllm_models: list[StoredModelInfo]

    @property
    def total_size_bytes(self) -> int:
        return sum(row.size_bytes for row in self.llamacpp_models + self.vllm_models)

    @property
    def total_models(self) -> int:
        return len(self.llamacpp_models) + len(self.vllm_models)
